# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import re
from msprobe.visualization.builder.msprobe_adapter import compare_node, get_compare_mode, run_real_data, get_csv_df, \
    get_api_indicator_info, get_real_api_data_list
from msprobe.visualization.utils import GraphConst, update_shared_dict, update_pbar_info
from msprobe.visualization.graph.graph import Graph, NodeOp
from msprobe.visualization.compare.mode_adapter import ModeAdapter
from msprobe.core.common.const import Const
from msprobe.core.compare.utils import get_compare_framework


class GraphComparator:
    MAX_DEPTH = 1000

    def __init__(self, graphs, dump_path_param, args, is_cross_framework, mapping_dict=None, pbar_info=None):
        self.graph_n = graphs[0]
        self.graph_b = graphs[1]
        self.dump_path_param = dump_path_param
        self.output_path = args.output_path
        self.parallel_merge = args.parallel_merge if hasattr(args, 'parallel_merge') else False
        self.ma = ModeAdapter(get_compare_mode(self.dump_path_param), self.parallel_merge)
        self.framework = get_compare_framework(dump_path_param.get("npu_path"), dump_path_param.get("bench_path"))
        self.layer_mapping = args.layer_mapping
        self.mapping_dict = mapping_dict
        self.fuzzy_match = args.fuzzy_match
        self.pattern = re.compile(r'\.\d+\.')
        self.is_cross_framework = is_cross_framework
        self.rank_pattern = re.compile(r"_rank\d+")
        self.pbar_info = pbar_info

    def compare(self):
        """
        比较函数，初始化结束后单独调用。比较结果写入graph_n
        """
        if self.fuzzy_match:
            self._compare_nodes_fuzzy(self.graph_n.root)
        else:
            self._compare_nodes(self.graph_n.root)
        self._postcompare()

    def add_compare_result_to_node(self, node, compare_result_list):
        """
        将比对结果添加到节点的输入输出数据中
        Args:
            node: 节点
            compare_result_list: 包含参数信息和对比指标（真实数据对比模式除外）的list
        """
        # 真实数据比对，先暂存节点，在多进程对比得到精度指标后，再将指标添加到节点中
        if self.ma.prepare_real_data(node):
            return
        api_indicator = get_api_indicator_info(self.ma.compare_mode, compare_result_list, self.ma.parallel_merge)
        compare_result_dict = {}
        # input和output对比数据分开
        for item in compare_result_list:
            if not isinstance(item, (list, tuple)) or not item:
                continue
            compare_result_dict[item[0]] = item
        precision_index = self.ma.parse_result(node, compare_result_dict, api_indicator)
        node.data[GraphConst.JSON_INDEX_KEY] = precision_index

    def _compare_nodes(self, node_root):
        """
        遍历NPU树中的节点，如果在Bench中找到具有相同名称的节点，检查他们的祖先和参数信息，检查一致则及逆行精度数据对比
        这里采用先序遍历，好处在于当这个节点被比较时，他的先序已经被匹配，这可以为后续的模糊匹配提供重要信息
        """

        def compare_single_node(node_n):
            if self.layer_mapping:
                node_b, ancestors_n, ancestors_b = Graph.mapping_match(node_n, self.graph_b, self.mapping_dict)
                if node_b:
                    ancestors_n.append(node_n.id)
                    ancestors_b.append(node_b.id)
                    node_n.matched_node_link = ancestors_b
                    node_b.matched_node_link = ancestors_n
            else:
                node_b, ancestors = Graph.match(self.graph_n, node_n, self.graph_b)
                if node_b:
                    ancestors.append(node_b.id)
                    node_n.add_link(node_b, ancestors)
            if node_b:
                # 真实数据比对只会得到基本信息，并没有精度指标，需要调用多进程对比接口
                self._get_and_add_result(node_n, node_b)
            node_list.extend(node_n.subnodes)

        if self.pbar_info:
            update_shared_dict(self.pbar_info.current_stage_dict, self.pbar_info.task_id, 1)
        total = 0
        node_list = [node_root]
        while node_list:
            compare_single_node(node_list.pop(0))
            total += 1
            if self.pbar_info:
                update_pbar_info(self.pbar_info, total, len(self.graph_n.node_map))

    def _compare_nodes_fuzzy(self, node_root):
        def compare_single_nodes_fuzzy(node_n):
            if node_n.op != NodeOp.function_api:
                # 模块经过模糊匹配
                node_b, ancestors_n, ancestors_b = Graph.fuzzy_match(node_n, self.graph_b.node_map.get(node_n.id))
                if node_b:
                    self._process_matched_nodes(node_n, node_b, ancestors_n, ancestors_b)
                    # 匹配上的两个模块中的所有api, 忽略dump调用次数，按照名称一致+模块中的调用顺序进行匹配
                    recount_result_n = self._recount_api_node(node_n)
                    recount_result_b = self._recount_api_node(node_b)
                    for recount_node_id, node_id_n in recount_result_n.items():
                        api_node_n = self.graph_n.node_map.get(node_id_n)
                        if not api_node_n:
                            continue
                        api_node_b, ancestors_n, ancestors_b = Graph.fuzzy_match(
                            api_node_n, self.graph_b.node_map.get(recount_result_b.get(recount_node_id)))
                        if api_node_b:
                            self._process_matched_nodes(api_node_n, api_node_b, ancestors_n, ancestors_b)
            node_list.extend(node_n.subnodes)

        if self.pbar_info:
            update_shared_dict(self.pbar_info.current_stage_dict, self.pbar_info.task_id, 1)
        total = 0
        node_list = [node_root]
        while node_list:
            compare_single_nodes_fuzzy(node_list.pop(0))
            total += 1
            if self.pbar_info:
                update_pbar_info(self.pbar_info, total, len(self.graph_n.node_map))

    def _postcompare(self):
        self._handle_api_collection_index()
        if not self.ma.compare_mode == GraphConst.REAL_DATA_COMPARE:
            return
        df = get_csv_df(True, self.ma.csv_data, self.ma.compare_mode)
        # 真实数据比对时，避免高频率进度监控线程与多进程池之间的GIL 抢占 + CPU 资源竞争，暂停进度监控线程，防止卡死
        if self.pbar_info and self.pbar_info.progress_dict is not None:
            self.pbar_info.set_wait_monitor(True)
        try:
            df = run_real_data(self.dump_path_param, df, self.framework, self.is_cross_framework)
        finally:
            if self.pbar_info and self.pbar_info.progress_dict is not None:
                self.pbar_info.set_continue_monitor(True)

        compare_data_dict = {row[0]: row.tolist() for _, row in df.iterrows()}
        for node in self.ma.compare_nodes:
            compare_result_list = get_real_api_data_list(node, compare_data_dict)
            api_indicator = get_api_indicator_info(self.ma.compare_mode, compare_result_list)
            precision_index = self.ma.parse_result(node, compare_data_dict, api_indicator)
            node.data[GraphConst.JSON_INDEX_KEY] = precision_index

    def _handle_api_collection_index(self):
        """
        api集合的指标, 使用集合中所有api最差的指标
        指标为1代表最差
        """

        def handle_api_collection_index(api_collection_node):
            precision_index = GraphConst.MIN_INDEX_KEY
            for api in api_collection_node.subnodes:
                precision_index = max(precision_index,
                                      api.data.get(GraphConst.JSON_INDEX_KEY, GraphConst.MIN_INDEX_KEY))
            api_collection_node.data[GraphConst.JSON_INDEX_KEY] = precision_index

        for node in self.graph_n.root.subnodes:
            if node.op == NodeOp.api_collection and node.id.startswith(GraphConst.APIS_BETWEEN_MODULES_ALL_RANKS):
                for sub_node in node.subnodes:
                    handle_api_collection_index(sub_node)
                handle_api_collection_index(node)
            elif node.op == NodeOp.api_collection:
                handle_api_collection_index(node)

    def _get_and_add_result(self, node_n, node_b):
        compare_result_list = compare_node(node_n, node_b, self.ma.compare_mode)
        if compare_result_list:
            self.ma.add_csv_data(compare_result_list)
            self.add_compare_result_to_node(node_n, compare_result_list)

    def _recount_api_node(self, node):
        """
        两个匹配上的模块, 忽略各自模块下所有api的dump调用次数, 并赋予模块中的调用顺序
        Return:
            {赋予模块中的调用顺序的node_id: 原始node_id}
        """
        recount_result = {}
        node_count = {}
        for sub_node in node.subnodes:
            if sub_node.op == NodeOp.function_api:
                # 忽略dump调用次数
                count_removed_id = self.pattern.sub(Const.SEP, sub_node.id)
                if self.rank_pattern.search(count_removed_id):
                    count_removed_id = self.rank_pattern.sub('', count_removed_id)
                node_count[count_removed_id] = node_count.get(count_removed_id, 0) + 1
                # 赋予模块中的调用顺序
                recount_node_id = count_removed_id + str(node_count.get(count_removed_id))
                recount_result[recount_node_id] = sub_node.id
        return recount_result

    def _process_matched_nodes(self, node_n, node_b, ancestors_n, ancestors_b):
        ancestors_n.append(node_n.id)
        ancestors_b.append(node_b.id)
        node_n.matched_node_link = ancestors_b
        node_b.matched_node_link = ancestors_n
        self._get_and_add_result(node_n, node_b)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import functools
import importlib
import types

import torch

from msprobe.core.common.log import logger
from msprobe.pytorch.common.utils import torch_version_above_or_equal_2
from msprobe.pytorch.dump.api_dump.api_register import get_api_register

if torch_version_above_or_equal_2:
    from torch._dynamo.convert_frame import convert_frame as _orig_convert_frame, Hooks


def wrap_jit_script_func():
    def patched_script(*args, **kwargs):
        all_api_registered = api_register.all_api_registered
        if all_api_registered:
            api_register.restore_all_api()
        result = original_script(*args, **kwargs)
        if all_api_registered:
            api_register.register_all_api()
        return result

    original_script = torch.jit.script
    api_register = get_api_register()
    torch.jit.script = patched_script


def wrap_compile_script_func():
    def _patched_convert_frame(compiler_fn, hooks):
        """
        在调用原 convert_frame 生成的 _convert_frame 之前恢复 API，
        调用完之后再重新注册所有 API。
        """
        # 拿到原来 inner 版的 _convert_frame
        inner_convert = _orig_convert_frame(compiler_fn, hooks)

        def _wrapped(frame: types.FrameType, cache_size: int, hooks: Hooks, frame_state):
            reg = get_api_register()
            # 进入前 restore
            reg.restore_all_api()
            try:
                result = inner_convert(frame, cache_size, hooks, frame_state)
            except Exception:
                # 异常时也要确保 register
                reg.register_all_api()
                raise
            # 正常结束后 register
            reg.register_all_api()
            return result

        # 保留原属性以兼容
        _wrapped._torchdynamo_orig_callable = compiler_fn  # type: ignore[attr-defined]

        _wrapped._clone_with_backend =\
            lambda backend: _patched_convert_frame(backend, hooks)  # type: ignore[attr-defined]

        return _wrapped

    import torch._dynamo.convert_frame as _cf_mod
    _cf_mod.convert_frame = _patched_convert_frame


def patch_dynamo_compile():
    cf = importlib.import_module("torch._dynamo.convert_frame")
    if not hasattr(cf, "_compile"):
        logger.warning("No found torch._dynamo.convert_frame._compile")

    original = cf._compile
    if getattr(original, "__msprobe_patched__", False):
        return

    @functools.wraps(original)
    def wrapped(*args, **kwargs):
        result = None
        try:
            reg = get_api_register()
            reg.restore_all_api()
        except Exception as e:
            logger.warning(f"[msprobe] Pre restore_all_api failed: {e}")
            return result

        try:
            result = original(*args, **kwargs)
        except Exception:
            logger.warning("[msprobe] _compile execution failed (returning None)")
            result = None
        finally:
            try:
                reg = get_api_register()
                reg.register_all_api()  # 改成注册hook
            except Exception as e:
                logger.warning(f"[msprobe] Post register_all_api failed: {e}")
        return result
    wrapped.__msprobe_patched__ = True
    wrapped.__msprobe_original__ = original
    cf._compile = wrapped


def unpatch_dynamo_compile() -> bool:
    # 预留取消patch接口
    cf = importlib.import_module("torch._dynamo.convert_frame")
    current = getattr(cf, "_compile", None)
    if current is None:
        return False
    original = getattr(current, "__msprobe_original__", None)
    if original is None:
        return False
    cf._compile = original
    return True


def preprocess_func():
    try:
        from torch.utils._device import _device_constructors
        _device_constructors()
    except ImportError:
        pass
    except Exception as e:
        logger.warning(f"Failed to execute _device_constructors. Error Details: {str(e)}")


def wrap_script_func():
    wrap_jit_script_func()
    if torch_version_above_or_equal_2:
        patch_dynamo_compile()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the tlv info parse function.

How to use:

1、Tag map:

TLV_CONFIG_TAB = [
    {'Name': 'data_type',  'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'data_id',    'TLV_Type': 'ATOM', 'Ele_Type': 'INT32'},
    {'Name': 'shape_dims', 'TLV_Type': 'TLNV', 'Ele_Type': 'UINT64', 'Tag': 1}
]

2、parse an bytes:
from msprobe.msaccucmp.cmp_utils.tlv_parse import TLV

class ParseInfo:
    def __init__(self):
        self.data_type = 0
        self.data_id = 0
        self.shape_dims = []

aux = b'0x5a\0x5a\0x5a\0x5a' + \
      b'0x11\0x11\0x11\0x11' + \
      b'0x01\0x00\0x00\0x00' + \
      b'0x10\0x00\0x00\0x00' + \
      b'0x0a\0x0a\0x0a\0x0b\0x0b\0x0b\0x0b\0x0b' + \
      b'0x0b\0x0b\0x0b\0x0b\0x0b\0x0b\0x0b\0x0b'

input = ParseInfo()
tlv = TLV(TLV_CONFIG_TAB)
aux, input = tlv.parse_tlv_by_cfg_tab(aux, input)

"""
import struct

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError

CHAR_SIZE = 1
INT16_SIZE = 2
UINT16_SIZE = 2
INT32_SIZE = 4
UINT32_SIZE = 4
INT64_SIZE = 8
UINT64_SIZE = 8

_UNPACK_FORMAT = {
    'CHAR': {'FMT': 's', 'SIZE': 1},
    'UINT16': {'FMT': 'H', 'SIZE': 2},
    'INT16': {'FMT': 'h', 'SIZE': 2},
    'UINT32': {'FMT': 'I', 'SIZE': 4},
    'INT32': {'FMT': 'i', 'SIZE': 4},
    'UINT64': {'FMT': 'Q', 'SIZE': 8},
    'INT64': {'FMT': 'q', 'SIZE': 8},
}

TLV_TYPE = ['ATOM', 'TLV', 'TLNV', 'NV']


class NestedObj:
    pass


class TLV:
    def __init__(self, tlv_tab, tag_type="UINT32", len_type="UINT32") -> None:
        self.tlv_tab = tlv_tab
        self.tag_type = tag_type
        self.len_type = len_type

    @staticmethod
    def _unpack_single_element(aux: bytes, ele_type) -> (bytes, any):
        type_para = _UNPACK_FORMAT.get(ele_type)
        ele_fmt, ele_size = type_para.get('FMT'), type_para.get('SIZE')

        if len(aux) < ele_size:
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)
        try:
            value = struct.unpack(ele_fmt, aux[:ele_size])[0]
        except struct.error as error:
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR) from error

        if ele_type == 'CHAR' and isinstance(value, bytes):
            value = value.decode(encoding="utf-8")

        return aux[ele_size:], value

    @staticmethod
    def _check_tl(tag, length, tlv_cfg):
        tag_in_cfg, ele_type = tlv_cfg.get("Tag"), tlv_cfg.get('Ele_Type')

        if tag_in_cfg != tag:
            log.print_error_log(
                'Failed to parse of tag_in_cfg: %s, tag_in_stream: %s.' % (tag_in_cfg, tag))
            raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)

        if not isinstance(ele_type, list):
            ele_size = _UNPACK_FORMAT.get(ele_type).get('SIZE')
            if length < ele_size:
                log.print_error_log(
                    'Failed to parse of tag: %s. length:%d < single ele_size:%d' % (tag_in_cfg, length, ele_size))
                raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)

    def parse_tlv_by_cfg_tab(self, content: bytes, x: object) -> (bytes, object):
        aux = content
        self._check_cfg_tab(self.tlv_tab)

        for tlv_cfg in self.tlv_tab:
            tlv_type = tlv_cfg.get('TLV_Type')
            if tlv_type == 'ATOM':
                aux, x = self._unpack_atom_value(aux, tlv_cfg, x)
            elif tlv_type == 'TLV':
                aux, x = self._unpack_tlv_value(aux, tlv_cfg, x)
            elif tlv_type == 'TLNV':
                aux, x = self._unpack_tlnv_value(aux, tlv_cfg, x)
            elif tlv_type == 'NV':
                num_attr_name = tlv_cfg.get('N')
                num = getattr(x, num_attr_name)
                aux, x = self._unpack_nv_value(aux, tlv_cfg, x, num)
        return aux, x

    def _check_cfg_tab(self, tlv_tab):
        for tlv_cfg in tlv_tab:
            tlv_type, ele_type = tlv_cfg.get('TLV_Type'), tlv_cfg.get('Ele_Type')
            if tlv_type not in TLV_TYPE:
                log.print_error_log('%s invalid tlv_type: %s.' % (tlv_cfg.get('Name'), tlv_type))
                raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)

            # Nested tlv
            if isinstance(ele_type, list):
                self._check_cfg_tab(ele_type)
            elif _UNPACK_FORMAT.get(ele_type) is None:
                log.print_error_log('invalid ele_type: %s.' % ele_type)
                raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)

            if tlv_type == 'NV':
                num_attr_name = tlv_cfg.get('N')
                if num_attr_name is None:
                    log.print_error_log("is tlv_tpye is NV, 'N' must be given!")
                    raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR)
        return

    def _unpack_atom_value(self, aux: bytes, tlv_cfg, x: object) -> (bytes, object):
        ele_type, ele_name = tlv_cfg.get('Ele_Type'), tlv_cfg.get('Name')

        if isinstance(ele_type, list):
            # parse Nested tlv value
            sub_tlv_tab = ele_type
            sub_tlv = TLV(sub_tlv_tab)
            nested_obj = NestedObj()
            aux, value = sub_tlv.parse_tlv_by_cfg_tab(aux, nested_obj)
        else:
            try:
                # parse single element
                aux, value = self._unpack_single_element(aux, ele_type)
            except CompareError as error:
                log.print_error_log('Failed to parse of ele name: %s, value_type: %s, '
                                    'Please check the dump file:%s.' % (tlv_cfg.get('Name'), ele_type, str(error)))
                raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR) from error

        setattr(x, ele_name, value)
        return aux, x

    def _parse_tag_len(self, aux: bytes) -> (bytes, any, any):
        aux, tag = self._unpack_single_element(aux, self.tag_type)
        aux, length = self._unpack_single_element(aux, self.len_type)
        return aux, tag, length

    def _unpack_tlv_value(self, aux: bytes, tlv_cfg, x: object) -> (bytes, object):
        ele_type, ele_name = tlv_cfg.get('Ele_Type'), tlv_cfg.get('Name')

        # tag, len
        try:
            aux_try, tag, length = self._parse_tag_len(aux)
        except CompareError:
            setattr(x, ele_name, None)
            log.print_info_log('Failed to parse of ele name: %s, ele_type: %s' % (ele_name, ele_type))
            return aux, x
        try:
            self._check_tl(tag, length, tlv_cfg)
        except CompareError:
            log.print_info_log('Failed to parse of ele name: %s, ele_type: %s' % (ele_name, ele_type))
            setattr(x, ele_name, None)
            return aux, x

        new_aux, new_x = self._unpack_atom_value(aux_try, tlv_cfg, x)
        return new_aux, new_x

    def _unpack_tlnv_value(self, aux: bytes, tlv_cfg, x: object) -> (bytes, object):
        tag_in_cfg, ele_type, ele_name = tlv_cfg.get("Tag"), tlv_cfg.get('Ele_Type'), tlv_cfg.get('Name')

        # tag, len
        try:
            aux_try, tag, length = self._parse_tag_len(aux)
        except CompareError:
            setattr(x, ele_name, None)
            log.print_info_log('skip to parse tlv of ele name: %s, ele_type: %s' % (ele_name, ele_type))
            return aux, x
        try:
            self._check_tl(tag, length, tlv_cfg)
        except CompareError:
            log.print_info_log('skip to parse tlv of ele name: %s, ele_type: %s' % (ele_name, ele_type))
            setattr(x, ele_name, None)
            return aux, x

        aux = aux_try

        # get list num of value
        if _UNPACK_FORMAT.get(ele_type) is not None:
            ele_size = _UNPACK_FORMAT.get(ele_type).get('SIZE')
            ele_num = int(length / ele_size)

            try:
                aux, x = self._unpack_nv_value(aux, tlv_cfg, x, ele_num)
            except CompareError as error:
                log.print_error_log('Failed to parse of ele name: %s, value_type: %s, '
                                    'Please check the dump file:%s.' % (tlv_cfg.get('Name'), ele_type, str(error)))
                raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR) from error

        elif isinstance(ele_type, list):
            value = []
            sub_aux = aux[:length]

            while (len(sub_aux) > 0):
                # Nested tlv
                sub_tlv_tab = ele_type
                sub_tlv, nested_obj = TLV(sub_tlv_tab), NestedObj()
                sub_aux, single_value = sub_tlv.parse_tlv_by_cfg_tab(sub_aux, nested_obj)
                value.append(single_value)

            setattr(x, tlv_cfg.get('Name'), value)
            aux = aux[length:]
        return aux, x

    def _unpack_nv_value(self, aux: bytes, tlv_cfg, x: object, ele_num) -> (bytes, object):
        ele_type, ele_name = tlv_cfg.get('Ele_Type'), tlv_cfg.get('Name')
        value = []

        for _ in range(ele_num):
            if isinstance(ele_type, list):
                # Nested tlv
                sub_tlv_tab = ele_type
                sub_tlv, nested_obj = TLV(sub_tlv_tab), NestedObj()
                aux, single_value = sub_tlv.parse_tlv_by_cfg_tab(aux, nested_obj)
                value.append(single_value)
            else:
                try:
                    # single element
                    aux, single_value = self._unpack_single_element(aux, ele_type)
                except CompareError as error:
                    log.print_error_log('Failed to parse of ele name: %s, value_type: %s, '
                                        'Please check the dump file:%s.' % (tlv_cfg.get('Name'), ele_type, str(error)))
                    raise CompareError(CompareError.MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR) from error
                value.append(single_value)

        if ele_type == 'CHAR' and isinstance(value, list):
            value_str = "".join(value)
            setattr(x, tlv_cfg.get('Name'), value_str)
        else:
            setattr(x, tlv_cfg.get('Name'), value)
        return aux, x



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import collections
from functools import reduce
import json

import numpy as np
from msprobe.msaccucmp.dump_parse.proto_dump_data import DumpData
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.conversion.dtype_conversion import hifloat8_to_float32, float8e4m3fn_to_float32, float8e5m2_to_float32


CommonAttr = collections.namedtuple('CommonAttr', ['data_type', 'tensor_format', 'address', 'original_shape'])


def _deserialize_dump_data_to_array(data, data_type, shape: list = None) -> any:
    """
    Deserialize dump data to array
    :param tensor: the dump data for input or output
    :return: the numpy array
    """
    if shape is not None and 0 in shape:
        return np.array([]).reshape(shape)
    if shape is not None:  # shape can be empty [] for scalar data
        cnt = 1
        for ii in shape:
            cnt *= ii
        cur_type = common.get_dtype_by_data_type(data_type)
        cur_byte = np.dtype(cur_type).itemsize
        result = np.frombuffer(data[:cur_byte * cnt], dtype=cur_type)
    else:
        result = np.frombuffer(data, dtype=common.get_dtype_by_data_type(data_type))
    if data_type in ConstManager.UNPACK_DTYPE:
        return np.unpackbits(result)
    elif data_type in ConstManager.CAST_FP32_DTYPE:
        return result.astype('float32')
    else:
        return result


def _deserialize_dump_data_fp_low_to_array(data, data_type, shape: list = None) -> any:
    """
    Deserialize the dump data in the unique special data format of the new chip
    @param data: the dump data for input or output
    @param data_type: the data target dtype
    @param shape: the data target shape
    @return: the numpy array
    """
    if shape is not None and 0 in shape:
        return np.array([]).reshape(shape)
    cur_type = common.get_dtype_by_data_type(data_type)
    if shape is not None:  # shape can be empty [] for scalar data
        cnt = 1
        for ii in shape:
            cnt *= ii
        cur_byte = 1
        if cur_type == "float8_e4m3fn":
            result = np.frombuffer(data[:cur_byte * cnt], dtype=np.uint8)
            result = np.array([float8e4m3fn_to_float32(ele) for ele in result])
        elif cur_type == "hifloat8":
            result = np.frombuffer(data[:cur_byte * cnt], dtype=np.uint8)
            result = np.array([hifloat8_to_float32(ele) for ele in result])
        elif cur_type == "float8_e5m2":
            result = np.frombuffer(data[:cur_byte * cnt], dtype=np.uint8)
            result = np.array([float8e5m2_to_float32(ele) for ele in result])
    else:
        result = np.frombuffer(data, dtype=np.uint8)
        if cur_type == "float8_e4m3fn":
            result = np.array(float8e4m3fn_to_float32(result))
        if cur_type == "hifloat8":
            result = np.array(hifloat8_to_float32(result))
        if cur_type == "float8_e5m2":
            result = np.array(float8e5m2_to_float32(result))          
    return result
 
 
def build_dump_tensor(dump_data_object_data: list, is_input: bool, is_ffts: bool) -> None:
    """
    replace the input or output object of DumpDataObj to DumpyTensor
    @param dump_data_object_data: input or output object of DumpDataObj
    @param is_input: if the tensor is input data
    @param is_ffts: if the tensor is ffts plus mode
    @return: None
    """
    for index, tensor in enumerate(dump_data_object_data):
        if not (hasattr(tensor, "shape") and tensor.shape) and tensor.size:
            log.print_info_log(f"Tensor shape is empty, using size {tensor.size} as shape")
            tensor.shape.append(tensor.size) # Ignore dtype size, just set a value large enough
        if tensor.data_type == DD.DT_UNDEFINED and tensor.size:
            tensor.shape.Clear()
            tensor.shape.append(tensor.size)
        if tensor.data_type in ConstManager.FP_LOW_DATA_DTYPE:
            data_to_np = _deserialize_dump_data_fp_low_to_array(tensor.data, tensor.data_type, list(tensor.shape.dim))
        else:
            data_to_np = _deserialize_dump_data_to_array(tensor.data, tensor.data_type, list(tensor.shape.dim))
        dump_tensor = DumpTensor(index, tensor.data_type, tensor.format, list(tensor.shape.dim),
                                 data_to_np, tensor.size, list(tensor.original_shape.dim),
                                 tensor.address, tensor.sub_format, is_input, is_ffts)
        dump_data_object_data[index] = dump_tensor


def build_nano_dump_tensor(dump_data_object_data: list, is_input: bool) -> None:
    """
    replace the input or output object of DumpDataObj to DumpyTensor
    @param dump_data_object_data: input or output object of DumpDataObj
    @param is_input: if the tensor is input data
    @param is_ffts: if the tensor is ffts plus mode
    @return: None
    """
    for index, tensor in enumerate(dump_data_object_data):
        data_to_np = _deserialize_dump_data_to_array(tensor.data, tensor.data_type, tensor.shape_dims)
        dump_tensor = DumpTensor(index, tensor.data_type, tensor.format, tensor.shape_dims,
                                 data_to_np, tensor.size, tensor.original_shape_dims,
                                 tensor.address, is_input=is_input, is_ffts=False)
        dump_data_object_data[index] = dump_tensor


class DumpTensor:
    """
    The class of DumpTensor, replace the class of DumpData.input or output.
    Include the data detail: index, data_type, tensor_format, shape, data, size, original_shape
    """

    def __init__(self: any, index: int = None, data_type: int = None, tensor_format: int = None,
                 shape: list = None, data: np.ndarray = None, size: int = None, original_shape: list = None,
                 address: int = None, sub_format: int = 0, is_input: bool = False, is_ffts: bool = False) -> None:

        self.index = index
        self.data_type = data_type
        self.tensor_format = tensor_format
        self.shape = shape if shape else []
        self.data = data
        self.size = size
        self.original_shape = original_shape
        self.address = address
        self.sub_format = sub_format
        self.is_input = is_input
        self.is_ffts = is_ffts

    @property
    def get_common_attr(self: any) -> tuple:
        """
        get common attr
        @return: tuple of common attr
        """
        common_attr = CommonAttr(self.data_type, self.tensor_format, self.address, self.original_shape)
        return common_attr


class DumpDataObj:
    """
    The class of DumpDataObject, replace the class DumpData or NanoDumpData.
    Include dump_file information
    """
    def __init__(self, dump_data=DumpData(), nano_dump_data=None) -> None:
        if nano_dump_data:
            self.version = nano_dump_data.version_id
            self.op_name = nano_dump_data.op_name
            self.dump_time = nano_dump_data.dump_time
            self.buffer = None
            self.space = None
            self.attr = None
            self.input_data = nano_dump_data.inputs
            self.output_data = nano_dump_data.outputs
            self.ffts_file_check = False

            build_nano_dump_tensor(self.output_data, is_input=False)
            build_nano_dump_tensor(self.input_data, is_input=True)

        else:
            self.version = dump_data.version
            self.op_name = dump_data.op_name
            self.dump_time = dump_data.dump_time
            self.buffer = dump_data.buffer
            self.space = [_space_data for _space_data in dump_data.space]
            self.attr = json.loads(dump_data.attr[0].value) if dump_data.attr else None
            self.input_data = [_input_data for _input_data in dump_data.input]
            self.output_data = [_output_data for _output_data in dump_data.output]

            is_ffts = False if self.get_ffts_mode is None else True
            build_dump_tensor(self.output_data, is_input=False, is_ffts=is_ffts)
            build_dump_tensor(self.input_data, is_input=True, is_ffts=is_ffts)

            self.ffts_file_check = True
            self.ffts_auto_input_shape_list = []
            self.ffts_auto_output_shape_list = []

    @property
    def get_dump_time(self: any) -> int:
        """
        get dump_time
        @return: dump time
        """
        return self.dump_time

    @property
    def get_thread_num(self: any) -> int:
        """
        get slice_instance_num
        @return: slice number
        """
        return self.attr.get("slice_instance_num")

    @property
    def get_cut_axis_manual(self: any) -> list:
        """
        calculate the cut axis of manual mode
        @return: cut axis
        """
        cut_axis = []
        if not self.attr or not self.attr["outputCutList"]:
            return cut_axis
        for output in self.attr["outputCutList"]:
            output_index = []
            for index, value in enumerate(output):
                if value != 1:
                    output_index.append(index)
            cut_axis.append(output_index)
        return cut_axis

    @property
    def get_cut_axis_auto(self: any) -> list:
        """
        calculate the cut axis of auto mode
        @return: cut axis
        """
        cut_axis = []
        if not self.attr or not self.attr["outputCutList"]:
            return cut_axis
        for output in self.attr["outputCutList"]:
            output_index = []
            for index, value in enumerate(output):
                if value != 1:
                    output_index.append(index)
            cut_axis.append(output_index)
        return cut_axis

    @property
    def get_ffts_mode(self: any) -> any:
        """
        get ffts+ mode
        @return: mode
        """
        return self.attr["threadMode"] if self.attr else None

    @staticmethod
    def check_shape_match(output_data: np.ndarray, shape: list) -> bool:
        if output_data.shape[-1] == reduce(lambda x, y: x * y, shape):
            return True
        else:
            log.print_error_log(
                f"The output_data shape {output_data.shape[-1]} doesn't match the shape in dump file {shape}")
            raise CompareError(CompareError.MSACCUCMP_UNMATCH_DATA_SHAPE_ERROR)

    def get_output_data(self: any) -> list:
        """
        Get output data
        @return: list of output data
        """
        output_data_list = []
        for output in self.output_data:
            if self.check_shape_match(output.data, output.shape):
                output_data_list.append(output.data.reshape(output.shape))
        return output_data_list

    def get_auto_output_data(self: any, output_shape: list) -> list:
        """
        Get output data
        @return: list of output data
        """
        output_data_list = []
        for index, output in enumerate(self.output_data):
            if self.check_shape_match(output.data, output_shape[index]):
                output_data_list.append(output.data.reshape(output_shape[index]))
        return output_data_list
    
    def calculate_auto_mode_shape(self: any, thread_id: int, tensor_type: str) -> list:
        """
        calculate the output data shape of auto mode
        @return: output shape
        """
        output_shape = []
        if tensor_type == ConstManager.INPUT:
            attr_name = "input_tensor_slice"
        elif tensor_type == ConstManager.OUTPUT:
            attr_name = "output_tensor_slice"
        if self.attr is not None and self.attr[attr_name]:
            for output in self.attr[attr_name][thread_id]:
                output_index = []
                for addr in output:
                    dim = addr.get("higher") - addr.get("lower")
                    output_index.append(dim)
                output_shape.append(output_index)
        return output_shape

    def set_op_attr(self: any, op_name: str, ffts_file_check: bool) -> None:
        """
        set op_name and ffts_file_check
        @param op_name: op name
        @param ffts_file_check: if file num doesn't match thread num
        @return: none
        """
        self.op_name = op_name
        self.ffts_file_check = ffts_file_check



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re
from abc import ABC, abstractmethod

from msprobe.core.config_check.utils.utils import config_checking_print
from msprobe.core.common.file_utils import FileOpen, load_yaml
from msprobe.core.common.const import Const, FileCheckConst


class Parser(ABC):
    @abstractmethod
    def parse(self, file_path: str) -> dict:
        pass

    def run(self, file_path: str) -> dict:
        """
            统一对外调用接口
        :param file_path: 需解析的文件路径
        :return:
        """
        try:
            result = self.parse(file_path)
        except Exception as exc:
            config_checking_print(f"{self.__class__} parsing error, skip file path: {file_path}, error: {exc}")
            result = {}
        return result


class ShellParser(Parser):
    def parse(self, file_path: str) -> dict:
        """
        Extracts arguments from bash script used to run a model training.
        """
        hyperparameters = {}
        script_content_list = []
        with FileOpen(file_path, 'r') as file:
            for line in file:
                stripped_line = line.lstrip()
                if not stripped_line.startswith('#'):
                    line = line.split('#')[0].rstrip() + '\n'
                    if line.strip():
                        script_content_list.append(line)
        script_content = ''.join(script_content_list)

        command_line = re.search(r'msrun\s[^|]*|torchrun\s[^|]*|python\d? -m torch.distributed.launch\s[^|]*',
                                 script_content,
                                 re.DOTALL)
        if command_line:
            command_line = command_line.group()

            blocks = re.findall(r'([a-zA-Z0-9_]{1,20}_ARGS)="(.*?)"', script_content, re.DOTALL)
            block_contents = {}
            for block_name, block_content in blocks:
                block_content = block_content.replace('\n', ' ')
                block_contents[block_name] = block_content
                command_line = command_line.replace(f"${block_name}", block_content)

            matches = re.findall(r'--([\w-]+)(?:\s+([^\s\\]+))?', command_line)
            for match in matches:
                key, value = match
                args_key = re.match(r'\$\{?(\w+)}?', value)
                if args_key:
                    env_vars = re.findall(rf'{args_key.group(1)}=\s*(.+)', script_content)
                    if env_vars:
                        value = env_vars[-1]
                hyperparameters[key] = value if value else True

        return hyperparameters


class YamlParser(Parser):
    hyperparameters = {}

    def parse(self, file_path: str) -> dict:
        ori_hyper = load_yaml(file_path)
        self.recursive_parse_parameters(ori_hyper, "")
        return self.hyperparameters

    def recursive_parse_parameters(self, parameters, prefix):
        if isinstance(parameters, dict):
            for key, value in parameters.items():
                new_prefix = prefix + Const.SEP + key if prefix else key
                self.recursive_parse_parameters(value, new_prefix)
        elif isinstance(parameters, list):
            if all(isinstance(x, (int, float, str, bool, list))for x in parameters):
                self.hyperparameters.update({prefix: parameters})
            else:
                for idx, value in enumerate(parameters):
                    new_prefix = prefix + Const.SEP + str(idx) if prefix else str(idx)
                    self.recursive_parse_parameters(value, new_prefix)
        elif isinstance(parameters, (int, float, str, bool)):
            self.hyperparameters.update({prefix: parameters})


class ParserFactory:
    __ParserDict = {
        FileCheckConst.SHELL_SUFFIX: ShellParser(),
        FileCheckConst.YAML_SUFFIX: YamlParser()
    }

    def get_parser(self, file_type: str) -> Parser:
        parser = self.__ParserDict.get(file_type, None)
        if not parser:
            raise ValueError(f'Invalid parser type: {file_type}')
        return parser



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from functools import wraps
from typing import Callable, Union, List
from enum import Enum


class EnumInstance(Enum):
    NO_INSTANCE = 0


class CheckResult:
    def __init__(self, passed=True, msg="") -> None:
        self.passed = bool(passed)
        self.msg = msg

    def __bool__(self):
        return self.passed

    def __str__(self) -> str:
        return self.get_msg()

    def __repr__(self) -> str:
        return f"{self.passed}:{self.msg}"

    def get_msg(self):
        return "pass" if self.passed else self.msg


class WaitingRule:
    def __init__(self, func, args, kwargs, err_msg=None) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.err_msg = f"{self.name} check failed." if err_msg is None else err_msg

    @property
    def name(self):
        return self.func.__name__

    def check(self, instence: "CheckerInstence"):
        passed = self.func(instence, *self.args, **self.kwargs)
        if isinstance(passed, CheckResult):
            return passed
        elif isinstance(passed, tuple) and len(passed) == 2:
            return CheckResult(passed[0], passed[1])
        else:
            return CheckResult(bool(passed), self.err_msg)


class Recorder:
    def __init__(self) -> None:
        self.record_set = set()

    def add_record(self, *names) -> None:
        self.record_set.update(names)

    def in_records(self, name) -> bool:
        return name in self.record_set

    def union(self, recorder: "Recorder"):
        self.record_set.update(recorder.record_set)


class RuleRunner:
    def __init__(self) -> None:
        self.waiting_rules: List[WaitingRule] = []
        self.default_rules: List[WaitingRule] = []
        self.running_default = False
        self.instence = None
        self.recorder = Recorder()

    def add_rule(self, waiting_rule):
        self.waiting_rules.append(waiting_rule)

    def set_recorder(self, recorder: Recorder):
        self.recorder = recorder

    def get_recorder(self) -> Recorder:
        return self.recorder

    def run_rule(self, waiting_rule: WaitingRule):
        if self.running_default and self.recorder.in_records(waiting_rule.name):
            return CheckResult()
        self.recorder.add_record(waiting_rule.name)
        return waiting_rule.check(self.instence)

    def implement_check(self, instence: "CheckerInstence"):
        self.instence = instence
        self.running_default = False

        for waiting_rule in self.waiting_rules:
            passed = self.run_rule(waiting_rule)
            if not passed:
                return passed

        self.running_default = True
        for waiting_rule in self.default_rules:
            passed = self.run_rule(waiting_rule)
            if not passed:
                return passed
        return CheckResult()

    def is_running(self):
        return self.instence is not None

    def as_default(self):
        self.default_rules = self.waiting_rules
        self.waiting_rules = []
        return self


class CheckerInstence:
    def __init__(self, instance=EnumInstance.NO_INSTANCE, converter=None):
        self.instance = instance
        self.converter = converter

    def has_instance(self):
        return self.instance != EnumInstance.NO_INSTANCE

    def get_instance(self):
        return self.instance

    def get_value(self):
        # for args type
        return self.instance

    def convert_instance(self):
        if self.converter is None:
            return CheckResult()

        self.instance, passed, err_msg = self.converter(self.instance)
        return CheckResult(passed, err_msg)


def rule(err_msg=None) -> Callable:
    if not isinstance(err_msg, (type(None), str)):
        raise TypeError("err_msg must be a string")

    def make_rule_wrapper(func: Callable) -> Callable:
        if not callable(func):
            raise ValueError("The @rule decorator can only be applied to functions")

        @wraps(func)
        def wrapper(checker: "Checker", *args, **kwargs) -> Union["Checker", CheckResult]:
            waiting_rule = WaitingRule(func, args, kwargs, err_msg)
            if checker.is_running():
                return checker.run_rule(waiting_rule)
            else:
                checker.add_rule(waiting_rule)

            if checker.has_instance():
                return checker.check(checker.get_instance())

            return checker

        return wrapper

    return make_rule_wrapper


class Checker(RuleRunner, CheckerInstence):
    def __init__(self, instance=EnumInstance.NO_INSTANCE, converter=None) -> None:
        RuleRunner.__init__(self)
        CheckerInstence.__init__(self, instance, converter)

    def check(self, value, will_raise=False) -> CheckResult:
        self.instance = value
        passed = self.convert_instance() and self.implement_check(self)
        if not passed and will_raise:
            raise ValueError(passed)
        return passed

    @rule(err_msg="value is None")
    def not_none(self) -> Union["Checker", CheckResult]:
        return self.instance is not None

    @rule(err_msg="value is not None")
    def is_none(self) -> Union["Checker", CheckResult]:
        return self.instance is None

    @rule()
    def not_eq(self, cmp_value) -> Union["Checker", CheckResult]:
        return self.instance != cmp_value, f"equal {cmp_value}"

    @rule()
    def eq(self, cmp_value) -> Union["Checker", CheckResult]:
        return self.instance == cmp_value, f"not equal {cmp_value}?"

    @rule()
    def is_type(self, *types) -> Union["Checker", CheckResult]:
        return self.not_none() and isinstance(self.instance, types), f"is not {types}"

    @rule()
    def any(self, *rules: "Checker") -> Union["Checker", CheckResult]:
        passed = CheckResult()
        for running_rule in rules:
            if self.running_default:
                running_rule.set_recorder(self.recorder)
            passed = running_rule.check(self.instance)
            if passed:
                if not self.running_default:
                    self.recorder.union(running_rule.get_recorder())
                return passed

        return passed

    @rule()
    def anti(self, anti_rule: "Checker") -> Union["Checker", CheckResult]:
        anti_rule.set_recorder(self.recorder)
        passed = anti_rule.check(self.instance)

        return not passed, f"not {str(passed)}"



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.dump.debugger.precision_debugger import BasePrecisionDebugger
from msprobe.pytorch.dump.debugger.debugger_config import DebuggerConfig
from msprobe.pytorch.dump.pt_config import parse_task_config
from msprobe.pytorch.dump.pytorch_service import PytorchService

from msprobe.core.common.const import Const, MsgConst
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.utils import check_token_range, ThreadSafe, check_rank_id
from msprobe.pytorch.common.utils import check_save_param, is_torch_nn_module
from msprobe.pytorch.dump.module_dump.module_dump import ModuleDumper


class PrecisionDebugger(BasePrecisionDebugger):

    def __init__(
        self,
        config_path=None,
        task=None,
        dump_path=None,
        level=None,
        step=None
    ):
        if self.initialized:
            return
        super().__init__(config_path, task, dump_path, level, step)
        self.config = DebuggerConfig(
            self.common_config,
            self.task_config,
            task,
            dump_path,
            level
        )
        self.service = PytorchService(self.config)
        self.module_dumper = ModuleDumper(self.service)
        self.ori_customer_func = {}

    @staticmethod
    def _get_task_config(task, json_config):
        return parse_task_config(task, json_config)

    @classmethod
    @ThreadSafe.synchronized
    def start(cls, model=None, token_range=None, rank_id=None):
        instance = cls._get_instance()
        if instance is None:
            return

        check_token_range(token_range)
        check_rank_id(rank_id)
        instance.config.check_model(model, token_range)
        instance.service.start(model, token_range, rank_id)

    @classmethod
    @ThreadSafe.synchronized
    def stop(cls):
        instance = cls._get_instance()
        if instance is None:
            return
        instance.service.stop()

    @classmethod
    @ThreadSafe.synchronized
    def step(cls):
        instance = cls._get_instance()
        if instance is None:
            return
        cls._instance.service.step()

    @classmethod
    @ThreadSafe.synchronized
    def save(cls, variable, name, save_backward=True):
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        if instance.task not in [Const.TENSOR, Const.STATISTICS] or instance.config.level != Const.LEVEL_DEBUG:
            return
        try:
            check_save_param(variable, name, save_backward)
        except ValueError:
            return
        instance.service.save(variable, name, save_backward)


@ThreadSafe.synchronized
def module_dump(module, dump_name):
    if not is_torch_nn_module(module):
        raise MsprobeException(
            MsprobeException.INVALID_PARAM_ERROR,
            f"the module argument in module_dump must be a torch.nn.Module type, "
            f"but currently there is an unsupported {type(module)} type."
        )
    if not isinstance(dump_name, str):
        raise MsprobeException(
            MsprobeException.INVALID_PARAM_ERROR,
            f"the dump_name argument in module_dump must be a str type"
        )
    instance = PrecisionDebugger._instance
    if not instance:
        raise MsprobeException(
            MsprobeException.INTERFACE_USAGE_ERROR,
            f"PrecisionDebugger must be instantiated before using module_dump interface"
        )
    instance.module_dumper.start_module_dump(module, dump_name)


@ThreadSafe.synchronized
def module_dump_end():
    instance = PrecisionDebugger._instance
    if not instance:
        raise MsprobeException(
            MsprobeException.INTERFACE_USAGE_ERROR,
            f"PrecisionDebugger must be instantiated before using module_dump_end interface"
        )
    instance.module_dumper.stop_module_dump()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import datetime
import os
import re
from collections import OrderedDict, defaultdict
from concurrent.futures import ProcessPoolExecutor, as_completed
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import pytz
from msprobe.core.common.const import MonitorConst
from msprobe.core.common.file_utils import (create_directory, read_csv,
                                            recursive_chmod, remove_path)
from msprobe.core.common.log import logger
from msprobe.core.common.utils import is_int
from msprobe.core.monitor.db_utils import MonitorDB, update_ordered_dict, get_ordered_list
from msprobe.core.monitor.utils import get_target_output_dir
from tqdm import tqdm


@dataclass
class CSV2DBConfig:
    """Configuration for CSV to database conversion"""
    monitor_path: str
    time_start: Optional[str] = None
    time_end: Optional[str] = None
    process_num: int = 1
    data_type_list: Optional[List[str]] = None
    output_dirpath: Optional[str] = None


def validate_process_num(process_num: int) -> None:
    """Validate process number parameter"""
    if not is_int(process_num) or process_num <= 0:
        raise ValueError("process_num must be a positive integer")
    if process_num > MonitorConst.MAX_PROCESS_NUM:
        raise ValueError(f"Maximum supported process_num is {MonitorConst.MAX_PROCESS_NUM}")


def validate_data_type_list(data_type_list: Optional[List[str]]) -> None:
    """Validate data type list parameter"""
    if data_type_list is None or not data_type_list:
        logger.info(f"Using default data types: {MonitorConst.METRICS_TRENDVIS_SUPPORTED}")
        return

    if not isinstance(data_type_list, list):
        raise ValueError("data_type_list must be a list")

    invalid_types = [t for t in data_type_list if t not in MonitorConst.METRICS_TRENDVIS_SUPPORTED]
    if invalid_types:
        raise ValueError(f"Unsupported data types: {invalid_types}")


def get_info_from_filename(file_name, metric_list=None):
    metric_name = "_".join(file_name.split('_')[:-1])
    if metric_list and metric_name not in metric_list:
        return "", 0, 0
    match = re.match(f"{metric_name}{MonitorConst.CSV_FILE_PATTERN}", file_name)
    if not match:
        return "", 0, 0
    step_start, step_end = match.groups()
    return metric_name, step_start, step_end


def _pre_scan_single_rank(rank: int, files: List[str]) -> Dict:
    """Pre-scan files for a single rank to collect metadata"""
    metrics = set()
    min_step = None
    max_step = 0
    metric_stats = defaultdict(set)
    targets = defaultdict(OrderedDict)

    for file_path in files:
        file_name = os.path.basename(file_path)
        metric_name, step_start, step_end = get_info_from_filename(file_name)
        if not metric_name:
            continue
        step_start, step_end = int(step_start), int(step_end)

        metrics.add(metric_name)
        min_step = min(
            step_start if min_step is None else min_step, step_start)
        max_step = max(max_step, step_end)

        data = read_csv(file_path)
        stats = [k for k in data.keys() if k in MonitorConst.OP_MONVIS_SUPPORTED]
        metric_stats[metric_name].update(stats)

        for row_id, row in data.iterrows():
            try:
                name = row[MonitorConst.HEADER_NAME]
                vpp_stage = int(row['vpp_stage'])
                micro_step = int(row.get('micro_step', MonitorConst.DEFAULT_INT_VALUE))
            except (ValueError, KeyError) as e:
                logger.warning(
                    f"CSV conversion failed | file={file_path}:{row_id+2} | error={str(e)}")
                continue
            target = (name, vpp_stage, micro_step)
            if target not in targets[metric_name]:
                targets[metric_name][target] = None

    return {
        'max_rank': int(rank),
        'min_step': min_step,
        'max_step': max_step,
        'metric_stats': metric_stats,
        'targets': {metric_name: list(target_dict.keys()) for metric_name, target_dict in targets.items()}
    }


def _pre_scan(monitor_db: MonitorDB, data_dirs: Dict[int, str], data_type_list: List[str], workers: int = 1):
    """Pre-scan all targets, metrics, and statistics"""
    logger.info("Scanning dimensions...")
    rank_files = defaultdict(list)

    # Collect files for each rank
    for rank, dir_path in data_dirs.items():
        files = os.listdir(dir_path)
        for file in files:
            metric_name, _, _ = get_info_from_filename(
                file, metric_list=data_type_list)
            if not metric_name:
                continue
            rank_files[rank].append(os.path.join(dir_path, file))

    # Parallel pre-scan
    with ProcessPoolExecutor(max_workers=workers) as executor:
        futures = {
            executor.submit(_pre_scan_single_rank, rank, files): rank
            for rank, files in rank_files.items()
        }

        results = []
        with tqdm(total=len(futures), desc="Pre-scanning ranks") as pbar:
            for future in as_completed(futures):
                rank = futures[future]
                try:
                    result = future.result()
                    results.append(result)
                except Exception as e:
                    logger.error(
                        f"Error pre-scanning rank {rank}: {str(e)}")
                pbar.update(1)

    # Aggregate results
    targets = defaultdict(OrderedDict)
    all_stats = set()
    global_stats = {
        "min_step": None,
        "max_step": 0,
        "max_rank": 0
    }
    for rank_result in results:
        global_stats["max_rank"] = max(global_stats["max_rank"], rank_result['max_rank'])
        if global_stats["min_step"] is None:
            global_stats["min_step"] = rank_result['min_step']
        else:
            global_stats["min_step"] = min(
                global_stats["min_step"],
                rank_result['min_step']
            )
        global_stats["max_step"] = max(global_stats["max_step"], rank_result['max_step'])

        for metric, stats in rank_result['metric_stats'].items():
            if metric not in global_stats:
                global_stats[metric] = set()
            global_stats[metric].update(stats)
            all_stats.update(stats)
        for metric_name in rank_result['targets']:
            targets[metric_name] = update_ordered_dict(
                targets[metric_name], rank_result['targets'][metric_name])

    monitor_db.insert_dimensions(targets)
    monitor_db.init_global_stats_data(global_stats)
    monitor_db.create_trend_data(
        get_ordered_list(all_stats, MonitorConst.OP_MONVIS_SUPPORTED))
    metric_id_dict = monitor_db.get_metric_mapping()
    target_dict = monitor_db.get_target_mapping()
    monitor_db.extract_tags_from_processed_targets(
        targets, metric_id_dict, target_dict)
    return rank_files, metric_id_dict, target_dict


def process_single_rank(
    task: Tuple[int, List[str]],
    metric_id_dict: Dict[str, Tuple[int, List[str]]],
    target_dict: Dict[Tuple[str, int, int], int],
    db_path: str
) -> int:
    """Process data import for a single rank"""
    rank, files = task
    db = MonitorDB(db_path)
    total_inserted = 0
    batch_data = []
    all_stats = get_ordered_list(
        set().union(*[stats for _, stats in metric_id_dict.values()]),
        MonitorConst.OP_MONVIS_SUPPORTED
    )
    for file in files:
        filename = os.path.basename(file)
        metric_name, _, _ = get_info_from_filename(filename)
        if not metric_name:
            continue
        metric_info = metric_id_dict.get(metric_name)
        if not metric_info:
            continue
        metric_id, _ = metric_info

        for row_id, row in read_csv(file).iterrows():
            try:
                # Parse row data
                name = row.get(MonitorConst.HEADER_NAME)
                vpp_stage = int(row['vpp_stage'])
                micro_step = int(row.get('micro_step', MonitorConst.DEFAULT_INT_VALUE))
                target_id = target_dict.get((name, vpp_stage, micro_step))
                if not target_id:
                    continue

                step = int(row['step'])
                # Prepare row data
                row_data = [rank, step, target_id, metric_id]
                row_data.extend(
                    float(row[stat]) if stat in row else None
                    for stat in all_stats
                )
            except (ValueError, KeyError) as e:
                logger.error(
                    f"CSV conversion failed | file={file}:{row_id+2} | error={str(e)}")
                continue

            batch_data.append(tuple(row_data))
            # Batch insert when threshold reached
            if len(batch_data) >= MonitorConst.BATCH_SIZE:
                inserted = db.insert_rows(batch_data)
                if inserted is not None:
                    total_inserted += inserted
                batch_data = []

    # Insert remaining data
    if batch_data:
        inserted = db.insert_rows(batch_data)
        if inserted is not None:
            total_inserted += inserted

    logger.info(f"Rank {rank} inserted {total_inserted} rows")
    return total_inserted


def import_data(monitor_db: MonitorDB, data_dirs: Dict[int, str], data_type_list: List[str], workers: int = 4) -> bool:
    """Main method to import data into database"""
    # 1. Pre-scan to get rank tasks
    monitor_db.init_schema()
    rank_tasks, metric_id_dict, target_dict = _pre_scan(monitor_db, data_dirs, data_type_list, workers)
    if not rank_tasks:
        logger.error("No valid data files found during pre-scan")
        return False

    # 2. Process data for each rank in parallel
    total_files = sum(len(files) for files in rank_tasks.values())
    logger.info(f"Starting data import for {len(rank_tasks)} ranks,"
                f"{total_files} files..."
                )
    all_succeeded = True
    with ProcessPoolExecutor(max_workers=workers) as executor:
        futures = {
            executor.submit(
                process_single_rank,
                (rank, files),
                metric_id_dict,
                target_dict,
                monitor_db.db_path): rank
            for rank, files in rank_tasks.items()
        }

        with tqdm(as_completed(futures), total=len(futures), desc="Import progress") as pbar:
            for future in pbar:
                rank = futures[future]
                try:
                    inserted = future.result()
                    pbar.set_postfix_str(
                        f"Rank {rank}: inserted {inserted} rows")
                except Exception as e:
                    logger.error(
                        f"Failed to process Rank {rank}: {str(e)}")
                    all_succeeded = False
    return all_succeeded


def csv2db(config: CSV2DBConfig) -> None:
    """Main function to convert CSV files to database"""
    validate_process_num(config.process_num)
    validate_data_type_list(config.data_type_list)

    target_output_dirs = get_target_output_dir(
        config.monitor_path, config.time_start, config.time_end)

    if config.output_dirpath is None:
        local_tz = pytz.timezone("Asia/Shanghai")
        cur_time = datetime.datetime.now(local_tz).strftime("%b%d_%H-%M-%S")
        config.output_dirpath = os.path.join(
            config.monitor_path, f"{cur_time}-csv2db")

    create_directory(config.output_dirpath)
    db_path = os.path.join(config.output_dirpath, "monitor_metrics.db")

    if os.path.exists(db_path):
        remove_path(db_path)
        logger.warning(f"Existing path {db_path} will be recovered")

    db = MonitorDB(db_path)

    result = import_data(
        db,
        target_output_dirs,
        config.data_type_list if config.data_type_list else MonitorConst.METRICS_TRENDVIS_SUPPORTED,
        workers=config.process_num
    )
    recursive_chmod(config.output_dirpath)
    if result:
        logger.info(
            f"Data import completed. Output saved to: {config.output_dirpath}")
    else:
        logger.warning(
            f"Data import may be incomplete. Output directory: {config.output_dirpath} "
            f"(Some records might have failed)"
        )



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import copy
import functools
import os
from abc import ABC, abstractmethod
from collections import defaultdict

from msprobe.core.common.exceptions import DistributedNotInitializedError
from msprobe.core.common.file_utils import create_directory
from msprobe.core.common.runtime import Runtime
from msprobe.core.common.utils import Const, print_tools_ends_info, DumpPathAggregation
from msprobe.core.dump.api_dump.api_registry import ApiRegistry
from msprobe.core.dump.data_dump.data_collector import build_data_collector
from msprobe.core.dump.kernel_dump.kernel_config import create_kernel_config_json
from msprobe.core.common.megatron_utils import MegatronStepInfo


class BaseService(ABC):
    def __init__(self, config):
        self.bench_dump_iter_dir = None
        self.config = copy.deepcopy(config)
        self.config.level = getattr(config, 'level_ori', config.level)  # 兼容MindSpore配置
        self.model = None
        self.data_collector = build_data_collector(self.config)
        self.current_iter = 0
        self.loop = 0
        self.init_step = 0
        self.cur_token_id = 0
        self.first_start = True
        self.primitive_switch = False
        self.current_rank = None
        self.dump_iter_dir = None
        self.should_stop_service = False
        self.hooked_modules = []
        self.ori_customer_func = {}
        self.debug_variable_counter = None
        self.current_step_first_debug_save = True
        self.logger = None  # 子类中注入
        self.api_register = None  # 子类中注入
        self.api_template = None  # 子类中注入
        self.hook_manager = None  # 子类中注入
        self._init_specific_components()
        self._register_api_hook()

    @property
    def _is_debug_level(self):
        return self.config.level == Const.LEVEL_DEBUG

    @property
    def _is_l2_level(self):
        return self.config.level == Const.LEVEL_L2

    @property
    def _is_mix_level(self):
        return self.config.level == Const.LEVEL_MIX

    @property
    def _is_need_module_hook(self):
        return self.config.level in [Const.LEVEL_MIX, Const.LEVEL_L0]

    @property
    def _is_need_api_hook(self):
        return self.config.level in [Const.LEVEL_MIX, Const.LEVEL_L1, Const.LEVEL_L2]

    @property
    def _is_no_dump_step(self):
        return (self.config.step and self.current_iter not in self.config.step)

    @property
    def _is_no_dump_rank(self):
        return (self.config.rank and self.current_rank not in self.config.rank)

    @property
    def _need_tensor_data(self):
        """判断是否需要采集tensor数据"""
        return bool(
            self.config.task in self.data_collector.tasks_need_tensor_data or
            (self.config.task == Const.STATISTICS and self.config.tensor_list)
        )

    @property
    @abstractmethod
    def _get_framework_type(self):
        """获取框架类型"""
        pass

    @staticmethod
    @abstractmethod
    def _get_current_rank():
        """获取当前rank_id"""
        pass

    @staticmethod
    def _change_jit_switch(status):
        """修改JitDump开关，mindspore子类重写"""
        pass

    def start(self, model=None, token_range=None, rank_id=None):
        """通用start模板"""
        self._process_iteration()
        if self._is_debug_level:
            return
        if model:
            self.model = model
        if self._is_need_module_hook and self.model not in self.hooked_modules:
            self._register_module_hook()
            self.hooked_modules.append(self.model)
        if self._need_stop_service():
            return
        Runtime.is_running = True
        self.cur_token_id = 0
        if self.first_start:
            if rank_id is not None:
                self.current_rank = rank_id
            else:
                try:
                    self.current_rank = self._get_current_rank()
                except DistributedNotInitializedError:
                    self.current_rank = None
            Runtime.current_rank = self.current_rank
            if self._is_no_dump_rank:
                Runtime.is_running = False
                return
            self._register_hook()
            self.first_start = False

            if token_range:
                self._register_infer_count_hook(self.model, token_range)
        self.logger.info(f"{Const.TOOL_NAME}: debugger.start() is set successfully")
        if token_range is None:
            self.primitive_switch = True
            self._change_jit_switch(True)
            self.logger.info(f"Dump switch is turned on at step {self.current_iter}. ")

        self.create_dirs()
        self.logger.info(f"Dump data will be saved in {self.dump_iter_dir}.")

    def stop(self):
        """通用stop模板"""
        if self._is_debug_level or self.should_stop_service:
            return
        if self._is_no_dump_step or self._is_no_dump_rank:
            return
        self.logger.info(f"{Const.TOOL_NAME}: debugger.stop() is set successfully. "
                         "Please set debugger.start() to turn on the dump switch again. ")
        Runtime.is_running = False
        self.primitive_switch = False
        self._change_jit_switch(False)
        if self._is_l2_level:
            return

        self._process_async_dump()
        self.data_collector.write_json()

    def step(self):
        """通用step处理"""
        if self.should_stop_service:
            return
        self._process_async_dump()
        self.data_collector.write_json()
        self.current_step_first_debug_save = True
        self.loop += 1
        self._reset_status()
        MegatronStepInfo.reset()

    def save(self, variable, name, save_backward):
        '''
        Args:
            variable: Union[List[variable], dict{str: variable}, mindspore.tensor, str, float, int]
            name: str
            save_backward: boolean
        Return:
            void
        '''
        if not self._is_debug_level:
            return
        self.current_iter = self.loop + self.init_step
        if self._is_no_dump_step:
            return

        if self.current_step_first_debug_save:
            try:
                self.current_rank = self._get_current_rank()
            except DistributedNotInitializedError:
                self.current_rank = None

            self.create_dirs()
            self.debug_variable_counter = defaultdict(int)
            self.current_step_first_debug_save = False

        count = self.debug_variable_counter[name]
        self.debug_variable_counter[name] += 1

        name_with_count = f"{name}.{count}"
        grad_name_with_count = f"{name}_grad.{count}"

        # forward save
        self.data_collector.debug_data_collect_forward(variable, name_with_count)

        # backward save
        if save_backward:
            self.data_collector.debug_data_collect_backward(variable, grad_name_with_count)

    def register_custom_api(self, module, api_name, api_prefix):
        self.ori_customer_func[str(module) + Const.SEP + api_name] = getattr(module, api_name)
        ApiRegistry.register_custom_api(module, api_name, api_prefix,
                                        functools.partial(self.build_hook, Const.API), self.api_template)

    def restore_custom_api(self, module, api):
        ori_func = self.ori_customer_func.get(str(module) + Const.SEP + api)
        if ori_func:
            setattr(module, api, ori_func)

    def build_hook(self, hook_type, name):
        return self.hook_manager.build_hook(hook_type, name)

    def create_dirs(self):
        """统一目录创建逻辑"""
        create_directory(self.config.dump_path)
        if Runtime.run_mode == Const.PYNATIVE_GRAPH_MODE:
            self.dump_iter_dir = os.path.join(self.config.dump_path, Const.PYNATIVE_MODE, f"step{self.current_iter}")
        else:
            self.dump_iter_dir = os.path.join(self.config.dump_path, f"step{self.current_iter}")

        if getattr(self.config, "bench_path", None):
            self.bench_dump_iter_dir = os.path.join(self.config.bench_path, f"step{self.current_iter}")
        else:
            self.bench_dump_iter_dir = None

        cur_rank = self.current_rank if self.current_rank is not None else ''
        if self._is_l2_level:
            self._create_l2_dirs(cur_rank)
        else:
            self._create_default_dirs(cur_rank)

    @abstractmethod
    def _init_specific_components(self):
        """初始化框架特定组件"""
        pass

    @abstractmethod
    def _register_hook(self):
        """注册hook函数"""
        pass

    @abstractmethod
    def _register_module_hook(self):
        """注册模块级别的hook函数"""

    def _need_stop_service(self):
        if self.should_stop_service:
            return True
        end_service = self.config.step and self.current_iter > max(self.config.step) or \
            self.data_collector and self.data_collector.data_processor.is_terminated
        if end_service:
            self.primitive_switch = False
            self._change_jit_switch(False)
            Runtime.is_running = False
            self.should_stop_service = True
            print_tools_ends_info()
            return True
        if self._is_no_dump_step:
            return True
        return False

    def _register_api_hook(self):
        if self._is_need_api_hook:
            self.api_register.initialize_hook(functools.partial(self.build_hook, Const.API))
            self.api_register.register_all_api()
            self.logger.info(f"The api {self.config.task} hook function is successfully mounted to the model.")

    def _register_infer_count_hook(self, root_model, token_range):
        """
        通过root_model执行的轮次来判断当前在第几个token
        param root_model: 需要采集的推理模型
        param token_range: [start, end], 采集infer的token循环范围，左右皆包含在内
        return: None
        """
        def infer_hook(model, args):
            if self.cur_token_id == token_range[0]:
                Runtime.is_running = True
                self.primitive_switch = True
                self._change_jit_switch(True)
                self.logger.info(f"Current token id: {self.cur_token_id}, start dump infer token.")
            elif token_range[0] < self.cur_token_id <= token_range[1]:
                self.logger.debug(f"Current token id: {self.cur_token_id}.")
            elif self.cur_token_id == token_range[1] + 1:
                Runtime.is_running = False
                self.primitive_switch = False
                self._change_jit_switch(False)
                self.logger.info(
                    f"Current token id: {self.cur_token_id}, exceed token_range, early stop dump infer token.")
            self.cur_token_id += 1
        # 此处root_model可以保证为 Module/Cell类型 或 [Module/Cell]类型
        if root_model and isinstance(root_model, list):
            root_model = root_model[0]
            self.logger.warning("Infer model can only input one to support token_range, choose the first one.")

        root_model.register_forward_pre_hook(infer_hook)

    def _create_l2_dirs(self, cur_rank):
        create_directory(self.dump_iter_dir)
        kernel_config_path = create_kernel_config_json(self.dump_iter_dir, cur_rank)
        self.config.kernel_config_path = kernel_config_path

    def _create_default_dirs(self, cur_rank):
        dump_dir = os.path.join(self.dump_iter_dir, f"rank{cur_rank}")
        bench_dump_dir = None
        if self.bench_dump_iter_dir:
            bench_dump_dir = os.path.join(self.bench_dump_iter_dir, f"rank{cur_rank}")
        create_directory(dump_dir)
        if bench_dump_dir:
            create_directory(bench_dump_dir)
        dump_data_dir = None
        if self._need_tensor_data:
            dump_data_dir = os.path.join(dump_dir, "dump_tensor_data")
            create_directory(dump_data_dir)

        self._configure_dump_paths(dump_dir, dump_data_dir, bench_dump_dir)


    def _configure_dump_paths(self, dump_dir, dump_data_dir, bench_dump_dir):
        dump_path_aggregation = DumpPathAggregation()
        dump_path_aggregation.dump_file_path = os.path.join(dump_dir, "dump.json")
        dump_path_aggregation.stack_file_path = os.path.join(dump_dir, "stack.json")
        dump_path_aggregation.construct_file_path = os.path.join(dump_dir, "construct.json")
        dump_path_aggregation.dump_error_info_path = os.path.join(dump_dir, "dump_error_info.log")
        dump_path_aggregation.dump_tensor_data_dir = dump_data_dir
        dump_path_aggregation.debug_file_path = os.path.join(dump_dir, "debug.json")
        if bench_dump_dir:
            dump_path_aggregation.bench_dump_file_path = os.path.join(bench_dump_dir, "dump.json")
        self.data_collector.update_dump_paths(dump_path_aggregation)
        self.data_collector.initialize_json_file(self._get_framework_type)

    def _process_iteration(self):
        """处理迭代计数"""
        self.current_iter = self.loop + self.init_step
        self.data_collector.update_iter(self.current_iter)
        Runtime.current_iter = self.current_iter

    def _process_async_dump(self):
        """处理异步dump逻辑"""
        if self.config.async_dump and self.config.task in [Const.STATISTICS, Const.TENSOR]:
            self.data_collector.data_processor.dump_async_data()

    def _reset_status(self):
        """通用状态重置"""
        self.data_collector.reset_status()
        self.hook_manager.reset_status()
        if self._is_l2_level:
            self.data_collector.data_processor.reset_status()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
FormatManager class.
This class mainly involves the execute_format_convert function.
"""
import os
import sys
import importlib
from functools import reduce

import numpy as np

from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class SrcToDest:
    """
    The class for src to dest
    """

    def __init__(self: any, src_format: int, dest_format: int, src_shape: any, dest_shape: any) -> None:
        self.src_shape = src_shape
        self.dest_shape = dest_shape
        self.src_format = src_format
        self.dest_format = dest_format

    def get_src_format(self: any) -> int:
        """
        Get src format
        :return: src format
        """
        return self.src_format

    def get_src_shape(self: any) -> any:
        """
        Get src shape
        :return: src shape
        """
        return self.src_shape


class FormatManager:
    """
    The class for format conversion manager
    """

    BUILT_IN_FORMAT_CONVERT_DIR_NAME = "builtin_format_convert"
    CUSTOM_FORMAT_CONVERT_DIR_NAME = "format_convert"
    CONVERT_FUNC_NAME = 'convert'
    CONVERT_ARG_COUNT = 3
    TO_FRACTAL_Z_FUNC_ARG_COUNT = 4

    def __init__(self: any, custom_path: str) -> None:
        self.custom_path = custom_path
        self.built_in_support_format = []
        self.custom_support_format = []
        self.support_format_map = {}

    @staticmethod
    def _add_format_file_to_list(file_path: str, support_format_list: list) -> bool:
        if os.path.isfile(file_path):
            _, match = RegManager.match_group(RegManager.FORMAT_CONVERT_FILE_NAME_PATTERN,
                                              os.path.basename(file_path))
            if match is not None:
                support_format_list.append(match.group(1))
                return True
            log.print_warn_log("The file '%s' does not match 'convert_{src_format}_to_{dest_format}.py in '%s', "
                               "please check the file." % (os.path.basename(file_path),
                                                           os.path.dirname(file_path)))
        return False

    def execute_format_convert(self: any, src_to_dest: SrcToDest, data: any, args: dict) -> np.ndarray:
        """
        Format convert
        :param src_to_dest: the format and shape for src and dest
        :param data: the data to convert
        :param args: the argument contain group conv
        :return: the array after conversion
        """
        src_format_str = common.get_format_string(src_to_dest.src_format)
        dest_format_str = common.get_format_string(src_to_dest.dest_format)
        module_name = 'convert_%s_to_%s' % (src_format_str, dest_format_str)
        if module_name not in self.support_format_map:
            log.print_warn_log("Not supported from %s to %s." % (src_format_str, dest_format_str))
            raise CompareError(CompareError.MSACCUCMP_INVALID_FORMAT_ERROR)

        # call convert function
        try:
            if args.get("group") < 2:
                new_array = self.support_format_map.get(module_name)(
                    src_to_dest.src_shape, src_to_dest.dest_shape, data)
            else:
                new_array = self.support_format_map.get(module_name)(
                    src_to_dest.src_shape, src_to_dest.dest_shape, data, args.get("group"))
        except Exception as err:
            log.print_error_log("Failed to execute '%s' in '%s'. %s"
                                % (self.CONVERT_FUNC_NAME, module_name, str(err)))
            raise CompareError(CompareError.MSACCUCMP_INVALID_CONVERT_FUNC_ERROR) from err
        finally:
            pass

        # check the return value is ndarray
        self._check_return_value_valid(new_array, module_name)

        return new_array

    def check_arguments_valid(self: any) -> None:
        """
        Check argument valid
        """
        self._make_support_format()
        self._make_support_format_map()

    def _make_support_format_by_path(self: any, dir_path: str, support_format_list: list) -> None:
        if not os.path.exists(dir_path):
            log.print_warn_log("There is no '%s' in '%s', please check the custom path."
                               % (self.CUSTOM_FORMAT_CONVERT_DIR_NAME, os.path.dirname(dir_path)))
            return
        one_match = False
        for item in os.listdir(dir_path):
            if self._add_format_file_to_list(os.path.join(dir_path, item), support_format_list):
                one_match = True
        if not one_match:
            log.print_warn_log("There is no legal 'convert_{src_format}_to_{dest_format}.py' file in '%s', "
                               "please check the path." % dir_path)

    def _make_support_format(self: any) -> None:
        """
        Make support format list for custom and built in
        """
        dir_path = os.path.join(os.path.dirname(__file__), self.BUILT_IN_FORMAT_CONVERT_DIR_NAME)
        self._make_support_format_by_path(dir_path, self.built_in_support_format)
        if self.custom_path:
            ret = path_check.check_path_valid(
                self.custom_path, True, False, path_check.PathType.Directory)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)

            path_check.check_path_all_file_exec_valid(self.custom_path)
            dir_path = os.path.join(self.custom_path, self.CUSTOM_FORMAT_CONVERT_DIR_NAME)
            self._make_support_format_by_path(dir_path, self.custom_support_format)
            sys.path.append(self.custom_path)

    def _get_module(self: any, format_name: str, dir_name: str, module_type: str) -> (bool, any):
        if dir_name is self.BUILT_IN_FORMAT_CONVERT_DIR_NAME:
            format_module = importlib.import_module('%s.%s.%s' % ("msprobe.msaccucmp.format_manager",
                                                                  dir_name, format_name))
        if dir_name is self.CUSTOM_FORMAT_CONVERT_DIR_NAME:
            format_module = importlib.import_module('%s.%s' % (dir_name, format_name))

        # check exist convert attr
        if not hasattr(format_module, self.CONVERT_FUNC_NAME):
            log.print_warn_log("[%s] The file '%s' has no attribute '%s'. Please check the file."
                               % (module_type, str(format_module.__file__), self.CONVERT_FUNC_NAME))
            return False, format_module
        return True, format_module

    def _get_function(self: any, format_module: any, module_type: str) -> (bool, any):
        format_func = getattr(format_module, self.CONVERT_FUNC_NAME)
        # check convert is function
        if not callable(format_func):
            log.print_warn_log("[%s] The '%s' in %s is not function. Please check the file." %
                               (module_type, self.CONVERT_FUNC_NAME, str(format_module.__file__)))
            return False, ''

        # check argument count of convert
        if format_func.__code__.co_argcount != self.CONVERT_ARG_COUNT \
                and 'to_FRACTAL_Z' not in format_module.__name__:
            log.print_warn_log("[%s] The argument count (%d) of '%s' in %s is not %d. Please check the file." %
                               (module_type, format_func.__code__.co_argcount, self.CONVERT_FUNC_NAME,
                                str(format_module.__file__), self.CONVERT_ARG_COUNT))
            return False, ''
        if format_func.__code__.co_argcount != self.TO_FRACTAL_Z_FUNC_ARG_COUNT \
                and 'to_FRACTAL_Z' in format_module.__name__:
            log.print_warn_log("[%s] The argument count (%d) of '%s' in %s is not %d. Please check the file." %
                               (module_type, format_func.__code__.co_argcount, self.CONVERT_FUNC_NAME,
                                str(format_module.__file__), self.TO_FRACTAL_Z_FUNC_ARG_COUNT))
            return False, ''
        return True, format_func

    def _make_support_format_map_by_list(self: any, format_list: list, dir_name: str, module_type: str) -> None:
        for format_name in format_list:
            get_ok, format_module = self._get_module(format_name, dir_name, module_type)
            if not get_ok:
                continue
            get_ok, format_func = self._get_function(format_module, module_type)
            if not get_ok:
                continue
            self.support_format_map[format_name] = format_func

    def _make_support_format_map(self: any) -> None:
        """
        Make support format map
        """
        self._make_support_format_map_by_list(self.built_in_support_format, self.BUILT_IN_FORMAT_CONVERT_DIR_NAME,
                                              ConstManager.BUILTIN)
        self._make_support_format_map_by_list(self.custom_support_format, self.CUSTOM_FORMAT_CONVERT_DIR_NAME,
                                              ConstManager.CUSTOM)
        if not self.support_format_map:
            log.print_error_log("There is no support format conversion.")
            raise CompareError(CompareError.MSACCUCMP_INVALID_CONVERT_FUNC_ERROR)

    def _check_return_value_valid(self: any, value: object, format_name: str) -> None:
        if not isinstance(value, np.ndarray):
            log.print_error_log(
                "The return value of '%s' in '%s' is not numpy.ndarray. Please check the return value."
                % (self.CONVERT_FUNC_NAME, format_name))
            raise CompareError(CompareError.MSACCUCMP_INVALID_CONVERT_FUNC_ERROR)


class ShapeConversion:
    """
    The class for shape conversion
    """

    def __init__(self: any, format_manager: FormatManager) -> None:
        self.format_manager = format_manager

    @staticmethod
    def reshape(src_to_dest: SrcToDest, array: any) -> any:
        """
        Reshape the array
        :param src_to_dest:the format and shape for src and dest
        :param array: the data array
        :return: the numpy array
        """
        shape = []
        size = 1
        for dim in src_to_dest.src_shape:
            shape.append(dim)
            size *= dim
        if 0 in array.shape:
            return array.reshape(shape)
        if size != len(array):
            log.print_error_log("The length(%d) is not match with the shape %s."
                                % (len(array), utils.convert_shape_to_string(src_to_dest.src_shape)))
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)
        return array.reshape(shape)

    @staticmethod
    def get_convert_fractal_nz_to_nd_dest_shape_length(src_to_dest):
        """
        Get convert FRACTAL_NZ to ND dest shape.
        :param src_to_dest: the format and shape for src and dest
        :return: the length of dest shape.
        """
        axis_h = reduce(lambda x, y: x * y, src_to_dest.dest_shape[:-2])
        axis_c1 = src_to_dest.src_shape[-4]
        axis_no = src_to_dest.src_shape[-3]
        axis_ni = src_to_dest.src_shape[-2]
        axis_c0 = src_to_dest.src_shape[-1]
        return axis_h * axis_c1 * axis_no * axis_ni * axis_c0

    def convert_shape(self: any, src_to_dest: SrcToDest, array: any, args: dict) -> any:
        """
        Check shape valid, then convert shape
        :param src_to_dest: the format and shape for src and dest
        :param array: the data to convert
        :param args: the argument contain group conv
        :return: the numpy array
        """
        # from format is ND, no need to convert
        if src_to_dest.src_format == DD.FORMAT_ND:
            return self.reshape(src_to_dest, array)
        # if format and shape is not match, return
        if src_to_dest.src_format == DD.FORMAT_NCHW \
                or src_to_dest.src_format == DD.FORMAT_HWCN \
                or src_to_dest.src_format == DD.FORMAT_NHWC:
            if len(src_to_dest.src_shape) != ConstManager.FOUR_DIMS_LENGTH:
                return self.reshape(src_to_dest, array)
        # src and dest format are the same, no need to convert
        if src_to_dest.src_format == src_to_dest.dest_format:
            return self.reshape(src_to_dest, array)
        # if convert FRACTAL_NZ to ND and the array is not equal to dest shape, no need to convert.
        if src_to_dest.src_format == DD.FORMAT_FRACTAL_NZ \
                and src_to_dest.dest_format == DD.FORMAT_ND \
                and len(src_to_dest.dest_shape) > 2:
            if len(array) != self.get_convert_fractal_nz_to_nd_dest_shape_length(src_to_dest):
                log.print_warn_log(
                    "Cannot convert FRACTAL_NZ to ND, the reason is the length "
                    "of array is not equal to the length of dest shape.")
                return self.reshape(src_to_dest, array)
        if src_to_dest.src_format == DD.FORMAT_FRACTAL_NZ:
            utils.check_shape_valid_in_nz(src_to_dest.dest_shape, src_to_dest.src_shape, is_convert_mode=False)
        return self.format_manager.execute_format_convert(src_to_dest, array, args)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import defaultdict

from msprobe.core.common.const import Data2DBConst
from msprobe.core.common.db_manager import DBManager, TrendSql
from msprobe.core.common.log import logger


def parse_full_key(full_key):
    """
    Module.blocks.3.layer.1.conv2.Conv2d.input.2
    """
    tags = set()
    parts = full_key.split('.')
    if len(parts) < 3:
        logger.debug(
            f"parse_full_key: key '{full_key}' has less than 3 parts,"
            f"returning empty set of tags")
        return set()
    # 关联index标签 full_key中最后一部分，如 'input.2'
    tags.add((".".join(parts[-2:]), Data2DBConst.TAG_INDEX))

    # 关联默认标签
    filtered_parts = []
    for part in parts[:-2]:
        if part not in Data2DBConst.DEFAULT_TAGS:
            filtered_parts.append(part)
        else:
            tags.add((part, Data2DBConst.TAG_DEFAULT))

    if not filtered_parts:
        return set()

    processed_indices = [-1]
    for i, part in enumerate(filtered_parts):
        if part.isdigit():
            if i == 0 or i - 1 in processed_indices:  # 确保数字前面有非数字部分, 组成layer标签
                processed_indices.append(i)
                continue
            layer_tag = (
                f"{filtered_parts[i-1]}.{filtered_parts[i]}", Data2DBConst.TAG_LAYER)
            tags.add(layer_tag)

            before_layer = ".".join(
                filtered_parts[processed_indices[-1] + 1: i - 1])
            if before_layer:
                module_tag = (before_layer, Data2DBConst.TAG_MODULE)
                tags.add(module_tag)
            processed_indices.append(i)

    if processed_indices[-1] != len(filtered_parts) - 1 and \
            ".".join(filtered_parts[processed_indices[-1] + 1:]):
        module_tag = (
            ".".join(filtered_parts[processed_indices[-1] + 1:]), Data2DBConst.TAG_MODULE)
        tags.add(module_tag)
    if len(processed_indices) == 1:
        api_tag = (".".join(filtered_parts), Data2DBConst.TAG_FUNCTION)
        tags.add(api_tag)
    return tags


class DumpDB:
    def __init__(self, db_path):
        self.db_path = db_path
        self.db_manager = DBManager(db_path)

        # 缓存
        self._metric_id_cache = {}  # metric_name -> metric_id
        self._stats_cache = {}      # metric_name -> stats list
        self._processed_targets = defaultdict(dict)  # 记录所有metric_id下的targets
        self._new_targets = defaultdict(list)  # 记录新增targets
        self._init_schema()
    
    def init_global_stats_data(self, config):
        """初始化全局统计表数据"""
        # 准备插入数据
        self.db_manager.execute_sql(
            TrendSql.create_global_stats_table(config)
        )
        
        column_values = []
        for _, column_value in config.items():
            if isinstance(column_value, (list, set)):
                column_values.append(",".join(column_value))
            elif isinstance(column_value, (int, float)):
                column_values.append(column_value)
        self.db_manager.insert_data("global_stats", [column_values,])

    def get_metric_id(self, metric_name: str):
        """Get metric ID by name"""
        if metric_name in self._metric_id_cache:
            return self._metric_id_cache[metric_name]
        result = self.db_manager.select_data(
            table_name="monitoring_metrics",
            columns=["metric_id"],
            where={"metric_name": metric_name}
        )
        metric_id = result[0]["metric_id"] if result else None
        self._metric_id_cache[metric_name] = metric_id
        return metric_id

    def cache_targets(self, target, metric_id):
        if target not in self._processed_targets[metric_id]:
            self._processed_targets[metric_id][target] = {'id': 0}
            self._new_targets[metric_id].append(target)
        return self._processed_targets[metric_id][target]

    def batch_insert_targets(self):
        """批量插入targets并通过rowid范围获取ID"""

        for metric_id in sorted(self._new_targets.keys()):
            targets = self._new_targets[metric_id]
            if not targets:
                continue
            self.db_manager.insert_data("monitoring_targets", list(targets), key_list=[
                                        "target_name", "vpp_stage", "micro_step"])
            results = self.db_manager.select_data(
                table_name="monitoring_targets",
                columns=["target_id", "target_name", "vpp_stage", "micro_step"]
            )
            # 更新缓存
            for row in results:
                cache_key = (row["target_name"],
                             row["vpp_stage"], row["micro_step"])
                if cache_key in self._processed_targets[metric_id]:
                    self._processed_targets[metric_id][cache_key]['id'] = row["target_id"]
        self._new_targets = defaultdict(list)

    def batch_insert_data(self, batch_data):
        """批量插入数据"""
        if not batch_data:
            return
        # 刷新target id
        for row in batch_data:
            if len(row) < 3 or not isinstance(row[2], dict):
                return
            row[2] = row[2]["id"]

        self.db_manager.insert_data("trend_data", batch_data)

    def extract_tags_from_processed_targets(self):
        """从已处理的targets中提取标签并建立映射关系"""

        for metric_id in self._processed_targets:
            tag_info_to_id = {}
            mapping_data = []
            target_to_tags = {}
            new_tags_data = set()
            for target, info in self._processed_targets[metric_id].items():
                target_name, _, _ = target
                target_id = info.get("id")

                if target_id is None:
                    continue
                # 从target_name中提取标签
                tags = parse_full_key(target_name)
                new_tags_data = new_tags_data.union(tags)
                target_to_tags[target_id] = tags

            # 插入新标签
            if not new_tags_data:
                continue
            self.db_manager.insert_data(
                "monitoring_tags",
                [(tag_name, category, metric_id)
                    for tag_name, category in new_tags_data],
                key_list=["tag_name", "category", "metric_id"]
            )

            # 重新获取tag_id用于映射
            tag_records = self.db_manager.select_data(
                "monitoring_tags",
                columns=["tag_id", "tag_name", "category", "metric_id"]
            )
            for row in tag_records:
                tag_info_to_id[(row["tag_name"], row["category"],
                                row["metric_id"])] = row["tag_id"]

            for target_id, tags_list in target_to_tags.items():
                for tag_name, category in tags_list:
                    tag_id = tag_info_to_id.get(
                        (tag_name, category, metric_id))
                    if tag_id:
                        mapping_data.append((target_id, tag_id))

            # 插入映射关系
            if mapping_data:
                self.db_manager.insert_data(
                    "tag_target_mapping",
                    mapping_data,
                    key_list=["target_id", "tag_id"]
                )

    def _init_schema(self) -> None:
        """Initialize database schema"""
        self.db_manager.execute_multi_sql(TrendSql.get_table_definition())
        self.db_manager.execute_sql(TrendSql.create_trend_table(Data2DBConst.ORDERED_STAT))
        for metric_name in Data2DBConst.METRICS:
            self.db_manager.insert_data("monitoring_metrics", [
                                        (metric_name,)], ["metric_name"])



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import sys
import threading
from collections import OrderedDict

import torch
from torch.utils.hooks import BackwardHook, RemovableHandle

from msprobe.core.common.const import Const
from msprobe.core.common.megatron_utils import wrap_megatron_step, get_micro_step, is_megatron
from msprobe.core.common.runtime import Runtime
from msprobe.core.common.utils import ModuleQueue, ThreadSafe
from msprobe.core.dump.data_dump.scope import BaseScope, ModuleRangeScope, MixRangeScope
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import is_torch_nn_module, register_forward_pre_hook
from msprobe.pytorch.dump.module_dump.hook_wrapper import (
    wrap_setup_input_output_hook,
    wrap_backward_hook_function_apply
)


torch_version_above_or_equal_2 = torch.__version__.split('+')[0] >= '2.0'
torch_version_above_or_equal_21 = torch.__version__.split('+')[0] >= '2.1'
if torch_version_above_or_equal_21:
    from torch.utils.checkpoint import _StopRecomputationError


def wrap_megatron_deallocate(func):
    def wrapper_func(out, deallocate_pipeline_outputs=False):
        if deallocate_pipeline_outputs and isinstance(out, torch.Tensor) and getattr(out, "_base") is not None:
            out_clone = out.clone()
            out.data = torch.empty((1,), device=out.device, dtype=out.dtype, )
            return func(out_clone, deallocate_pipeline_outputs)
        return func(out, deallocate_pipeline_outputs)

    return wrapper_func


def wrap_forward_with_hook_safety(module):
    """
    包装模块的forward方法，确保异常时也执行forward_hook。
    """
    original_forward = module.forward

    def wrapped_forward(*args, **kwargs):
        try:
            output = original_forward(*args, **kwargs)
            return output
        except _StopRecomputationError as e:
            exception_output = None
            if len(module._forward_hooks.values()) > 0:
                # 仅执行msprobe的forward_hook, hook名称必然包含'ModuleProcessor.'
                for hook_fn in module._forward_hooks.values():
                    if 'ModuleProcessor' in str(hook_fn):
                        hook_fn(module, args, kwargs, exception_output)
                        break
            raise e

    if torch_version_above_or_equal_21:
        module.forward = wrapped_forward


class ModuleProcessor:
    module_queue = ModuleQueue()
    module_count = {}
    module_stack = {}
    api_parent_node = {}
    module_node = {}
    module_bw_hook_kernels = {}
    module_with_backward_hook = {}
    enable_module_dump = False
    is_megatron_module = False

    def __init__(self, scope):
        self.scope = scope if isinstance(scope, (ModuleRangeScope, MixRangeScope)) else None
        wrap_setup_input_output_hook()
        wrap_backward_hook_function_apply()
        try:
            from megatron.core.pipeline_parallel import schedules
            origin_func_id = id(schedules.deallocate_output_tensor)
            schedules.deallocate_output_tensor = wrap_megatron_deallocate(schedules.deallocate_output_tensor)
            schedules.forward_step = wrap_megatron_step(schedules.forward_step)
            schedules.backward_step = wrap_megatron_step(schedules.backward_step, is_forward=False)
            for module in list(sys.modules.values()):
                if module.__name__ == 'schedules':
                    continue
                for func in module.__dict__:
                    if id(module.__dict__[func]) == origin_func_id:
                        module.__setattr__(func, schedules.deallocate_output_tensor)
                        logger.debug(f'patch {module.__name__}.{func}.')
            logger.info_on_rank_0("Patch megatron method success.")
        except ImportError:
            logger.info_on_rank_0("No megatron find.")
        except Exception as e:
            logger.info_on_rank_0(f"Patch megatron method failed, detail:{str(e)}")

    @staticmethod
    def set_and_get_calls_number(module_name):
        if module_name not in ModuleProcessor.module_count:
            ModuleProcessor.module_count[module_name] = 0
        else:
            ModuleProcessor.module_count[module_name] += 1
        return ModuleProcessor.module_count[module_name]

    @staticmethod
    def has_register_backward_hook(module):
        return hasattr(module, '_backward_hooks') and \
            len(module._backward_hooks) > 0 and \
            module._is_full_backward_hook is False

    @staticmethod
    def get_modules_and_names(models, recursive, module_names):
        modules_and_names_with_index = {}
        if isinstance(models, (list, tuple)):
            if not recursive and len(module_names) != len(models):
                return modules_and_names_with_index
            for index, model in enumerate(models):
                modules_and_names_with_index[str(index)] = model.named_modules() if recursive else \
                    [(module_names[index], model)]
        else:
            if not recursive and len(module_names) != 1:
                return modules_and_names_with_index
            modules_and_names_with_index["-1"] = models.named_modules() if recursive else \
                [(module_names[0], models)]
        return modules_and_names_with_index

    @staticmethod
    def check_is_megatron_module(models):
        if isinstance(models, (list, tuple)):
            for model in models:
                if "megatron" in type(model).__module__:
                    return True
        else:
            return "megatron" in type(models).__module__

    @classmethod
    def reset_module_stats(cls):
        cls.module_queue = ModuleQueue()
        cls.module_count = {}
        cls.module_stack = {}
        cls.api_parent_node = {}
        cls.module_node = {}
        cls.module_bw_hook_kernels = {}
        cls.enable_module_dump = False

    def register_module_hook(self, models, build_hook, recursive=True, module_names=None):
        if module_names is None:
            module_names = []

        ModuleProcessor.is_megatron_module = self.check_is_megatron_module(models)
        modules_and_names_with_index = self.get_modules_and_names(models, recursive, module_names)
        for index, modules_and_names in modules_and_names_with_index.items():
            model = models if index == "-1" else models[int(index)]
            
            model_list = []
            for name, module in modules_and_names:
                model_list.append((name, module))
            
            is_verl = "verl" in sys.modules
            for idx, (name, module) in enumerate(model_list):
                if recursive and module == model:
                    continue
                if not is_torch_nn_module(module):
                    logger.warning(
                        f"The module dump does not support {type(module)} type. "
                        f"The data dump for this module will be skipped."
                    )
                    continue
                if module.__class__.__name__ == "FullyShardedDataParallel":
                    continue
                
                # verl 场景下跳过第一层和最后一层
                if is_verl and (idx == 1 or idx == len(model_list) - 1):
                    logger.warning(f"The module {name} is the first or last layer in verl scenario, "
                                   f"the data dump for this module will be skipped.")
                    continue
                
                setattr(module, 'msprobe_hook', True)
                module_index = (index + Const.SEP) if index != "-1" else ""
                prefix_name = f'{BaseScope.Module_Type_Module}{Const.SEP}{module_index}{name}{Const.SEP}' + \
                              f'{module.__class__.__name__}{Const.SEP}'

                forward_pre_hook = self.build_module_hook(prefix_name, build_hook)

                if self.has_register_backward_hook(module):
                    logger.warning(
                        f"The {prefix_name[:-1]} has registered deprecated register_backward_hook,"
                        f"which may cause abnormal data dump. The backward data dump for this module will be skipped."
                    )
                    ModuleProcessor.module_with_backward_hook[prefix_name] = True
                wrap_forward_with_hook_safety(module)
                register_forward_pre_hook(module, forward_pre_hook)

    def build_module_hook(self, module_name, build_data_hook):
        @ThreadSafe.synchronized
        def forward_pre_hook(module, args, kwargs=None):
            if kwargs is None:
                kwargs = {}

            if not Runtime.is_running:
                return (args, kwargs) if torch_version_above_or_equal_2 else args

            if hasattr(module, 'msprobe_module_dump') and not self.enable_module_dump:
                return (args, kwargs) if torch_version_above_or_equal_2 else args

            index = ModuleProcessor.set_and_get_calls_number(module_name)
            full_forward_name = f'{module_name}{Const.FORWARD}{Const.SEP}{index}'
            full_backward_name = f'{module_name}{Const.BACKWARD}{Const.SEP}{index}'

            self.set_construct_info_in_pre_hook(full_forward_name)

            if not hasattr(module, 'msprobe_forward_hook'):
                forward_hooks_dict = getattr(module, '_forward_hooks', OrderedDict())
                handle = RemovableHandle(forward_hooks_dict)
                forward_hooks_dict[handle.id] = forward_hook
                forward_hooks_dict.move_to_end(handle.id, last=False)
                if torch_version_above_or_equal_2:
                    forward_hooks_with_kwargs_dict = getattr(module, '_forward_hooks_with_kwargs', OrderedDict())
                    forward_hooks_with_kwargs_dict[handle.id] = True

                setattr(module, 'msprobe_forward_hook', True)

            hook_set = build_data_hook(BaseScope.Module_Type_Module, full_forward_name)

            def get_backward_pre_hook(full_backward_name):
                @ThreadSafe.synchronized
                def backward_pre_hook_fn(module, grad_output):
                    self.set_construct_info_in_pre_hook(full_backward_name)

                return backward_pre_hook_fn

            def get_backward_hook(backward_data_hook, full_backward_name):
                @ThreadSafe.synchronized
                def backward_hook_fn(module, grad_input, grad_output):
                    new_output = backward_data_hook(module, grad_input, grad_output)
                    self.set_construct_info_in_hook(full_backward_name, is_forward=False)
                    return new_output

                return backward_hook_fn

            if not ModuleProcessor.module_with_backward_hook.get(module_name):
                backward_pre_hook = get_backward_pre_hook(full_backward_name)
                backward_hook = get_backward_hook(hook_set.backward_hook, full_backward_name)
                if torch_version_above_or_equal_2:
                    bw_hook = BackwardHook(module, [backward_hook], [backward_pre_hook])
                else:
                    bw_hook = BackwardHook(module, [backward_hook])
                ModuleProcessor.module_bw_hook_kernels[full_forward_name] = bw_hook
                args = bw_hook.setup_input_hook(args)
            return (args, kwargs) if torch_version_above_or_equal_2 else args

        @ThreadSafe.synchronized
        def forward_hook(module, args, kwargs_or_output, output_or_kwargs=None):
            if hasattr(module, 'msprobe_module_dump') and not self.enable_module_dump:
                return output_or_kwargs if torch_version_above_or_equal_2 else kwargs_or_output

            index = ModuleProcessor.module_count.get(module_name)
            full_name = f'{module_name}{Const.FORWARD}{Const.SEP}{index}'

            hook_set = build_data_hook(BaseScope.Module_Type_Module, full_name)
            hook_result = hook_set.forward_hook(module, args, kwargs_or_output, output_or_kwargs)
            self.set_construct_info_in_hook(full_name)

            if hook_result is not None:
                result = hook_result
            else:
                result = output_or_kwargs if torch_version_above_or_equal_2 else kwargs_or_output

            bw_hook = ModuleProcessor.module_bw_hook_kernels.get(full_name)
            if bw_hook:
                result = bw_hook.setup_output_hook(result)

            return result

        return forward_pre_hook

    def set_construct_info_in_pre_hook(self, full_name):
        tid = threading.get_ident()
        if tid not in self.module_stack:
            ModuleProcessor.module_stack[tid] = []

        if self.module_stack[tid]:
            ModuleProcessor.module_node[full_name] = self.module_stack[tid][-1] if not is_megatron() \
                else [self.module_stack[tid][-1], get_micro_step()]
        else:
            parent_name = ModuleProcessor.module_queue.find_last(full_name)
            ModuleProcessor.module_node[full_name] = parent_name if not is_megatron() \
                else [parent_name, get_micro_step()]

        ModuleProcessor.module_queue.add_name(full_name)
        ModuleProcessor.module_stack[tid].append(full_name)
        ModuleProcessor.api_parent_node[tid] = full_name if not is_megatron() else [full_name, get_micro_step()]
        if self.scope:
            self.scope.begin_module(full_name)

    def set_construct_info_in_hook(self, full_name, is_forward=True):
        tid = threading.get_ident()
        if torch_version_above_or_equal_2 or is_forward:
            ModuleProcessor.module_queue.remove_name(full_name)
            ModuleProcessor.api_parent_node[tid] = None if not is_megatron() else [None, get_micro_step()]
            if self.module_stack.get(tid):
                ModuleProcessor.module_stack[tid].pop()
            if self.module_stack.get(tid):
                ModuleProcessor.api_parent_node[tid] = ModuleProcessor.module_stack[tid][-1] if not is_megatron() \
                    else [ModuleProcessor.module_stack[tid][-1], get_micro_step()]
            if self.scope:
                self.scope.end_module(full_name)
        else:
            if self.scope:
                self.scope.begin_module(full_name)
            ModuleProcessor.api_parent_node[tid] = full_name if not is_megatron() else [full_name, get_micro_step()]



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from abc import abstractmethod

import torch

from msprobe.pytorch.common.log import logger
from msprobe.core.monitor.utils import MVResult
from msprobe.pytorch.monitor.module_metric import get_metrics
from msprobe.core.common.const import MonitorConst


class OptimizerMon(object):
    def __init__(self, torch_opt) -> None:
        self.fp16_to_fp32_param = {}
        self.torch_opt = torch_opt
        self.state = {}
        self.origin_funcs = []
        self.bucket_class = None

    def narrow_from_flatten(self, param, flatten_state):
        return flatten_state

    def get_state(self, torch_opt):
        if hasattr(torch_opt, 'chained_optimizers'):
            for opt in torch_opt.chained_optimizers:
                self._get_single_state(opt)
        else:
            self._get_single_state(torch_opt)

    def fetch_grad(self, monitor, params2name):
        if not self.fp16_to_fp32_param:
            self.map_fp16_to_fp32_param(self.torch_opt)

        grad_dict = {}
        first_param = True
        for param, name in params2name.items():
            if monitor.duplicate_param.get(name, False):
                continue
            if self.fp16_to_fp32_param and param not in self.fp16_to_fp32_param:
                continue
            grad = param.main_grad if monitor.params_have_main_grad else param.grad
            if grad.__class__.__name__ == 'DTensor':
                grad = grad.to_local()
            element_in_cur_partition = self.fp16_to_fp32_param.get(param, param).numel()
            if param.numel() != element_in_cur_partition:
                if first_param:
                    grad = grad.flatten()[-element_in_cur_partition:]
                else:  # supposed to be the last one
                    grad = grad.flatten()[:element_in_cur_partition]
            first_param = False

            if grad is None:
                if not monitor.fsdp_wrapped_module:
                    logger.warning(f"grad is None: {name}, maybe something wrong happened.")
                continue
            tag = monitor.name2tag.get(name, {}).get(MonitorConst.POST_GRAD)
            monitor.register_param_call_id("hook_optimizer", tag)
            grad_dict[tag] = grad
        return grad_dict

    def map_fp16_to_fp32_param(self, torch_opt):
        pass

    def fetch_mv(self, monitor, params2name):
        if not self.fp16_to_fp32_param:
            self.map_fp16_to_fp32_param(self.torch_opt)
        if not self.state:
            self.get_state(self.torch_opt)

        exp_avg_dict = {}
        exp_avg_sq_dict = {}
        update_dict = {}
        ratio_dict = {}

        if not self.state:
            logger.warning('optimizer state can not accessed')
            return MVResult(exp_avg=exp_avg_dict, exp_avg_sq=exp_avg_sq_dict, update=update_dict, ratio=ratio_dict)

        for lp_param, name in params2name.items():
            if lp_param in self.fp16_to_fp32_param:
                hp_param = self.fp16_to_fp32_param[lp_param]
            else:
                hp_param = lp_param

            if hp_param in self.state:
                state_param = self.state.get(hp_param, {})
                exp_avg = self.narrow_from_flatten(lp_param, state_param.get("exp_avg", None))
                exp_avg_sq = self.narrow_from_flatten(lp_param, state_param.get("exp_avg_sq", None))
                if monitor.mv_distribution:
                    exp_avg_dict[name] = exp_avg
                    exp_avg_sq_dict[name] = exp_avg_sq
                if monitor.mg_direction:
                    exp_avg_dict[name] = exp_avg
                if monitor.ur_distribution:
                    if len(self.torch_opt.param_groups) > 1:
                        logger.info(f"the length of torch_opt.param_groups is {len(self.torch_opt.param_groups)}.")
                    if 'step' in state_param:
                        step = state_param['step']  # Optimizer from pytorch or FusedAdam from apex(used by megatron)
                    elif 'step' in self.torch_opt.param_groups[0]:
                        step = self.torch_opt.param_groups[0]['step']  # AdamW from mindspeed
                    else:
                        logger.warning(f"step of {name} is None, maybe something wrong happened.")
                        continue
                    if exp_avg is None or exp_avg_sq is None:
                        logger.warning(f"exp_avg or exp_avg_sq of {name} is None, skip calculation.")
                        continue
                    exp_avg_hat = exp_avg / (1 - self.torch_opt.defaults['betas'][0] ** step)
                    exp_avg_sq_hat = exp_avg_sq / (1 - self.torch_opt.defaults['betas'][1] ** step)
                    update_dict[name] = exp_avg_hat / (torch.sqrt(exp_avg_sq_hat) + self.torch_opt.defaults['eps'])
                    ratio_dict[name] = exp_avg_hat / torch.sqrt(exp_avg_sq_hat)
                    monitor.update_heatmap_visualizer[name].pre_cal(update_dict[name])
                    monitor.ratio_heatmap_visualizer[name].pre_cal(ratio_dict[name])
        return MVResult(exp_avg=exp_avg_dict, exp_avg_sq=exp_avg_sq_dict, update=update_dict, ratio=ratio_dict)

    def patch_grad_sync(self, monitor):
        def patch_sync(sync_grad_func):
            def wrapper(bucket):
                grad_dict = {}
                # Megatron between core_r0.6.0 and core_r0.8.0, this bucket is Bucket.
                # When megatron is core_r0.9.0, this bucket is _ParamAndGradBucketGroup.
                # In megatron version core_r0.9.0, func start_grad_sync from Bucket moved to _ParamAndGradBucketGroup.
                bucket_params_id_list = [id(params) for params in bucket.params]
                for param, name in monitor.param2name.items():
                    if id(param) not in bucket_params_id_list:
                        continue
                    grad = param.main_grad if monitor.params_have_main_grad else param.grad
                    if grad is None:
                        logger.warning(f"grad is None: {name}, maybe something wrong happened.")
                        continue
                    tag = monitor.name2tag.get(name, {}).get(MonitorConst.PRE_GRAD)
                    if tag is None:
                        continue
                    grad_dict[tag] = grad
                    monitor.register_param_call_id("sync_grad_func", tag)
                get_metrics(monitor.ops, grad_dict, monitor.eps, monitor.grad_context.pre)
                out = sync_grad_func(bucket)
                return out

            return wrapper

        try:
            from megatron.core.distributed.param_and_grad_buffer import Bucket
            self.origin_funcs.append(Bucket.start_grad_sync)
            self.bucket_class = Bucket
            Bucket.start_grad_sync = patch_sync(Bucket.start_grad_sync)
            monitor.enable_megatron = True
            logger.info("megatron version is >= core_r0.6.0 <= core_r0.8.0")
        except ImportError:
            monitor.enable_megatron = False

        try:
            from megatron.core.distributed.param_and_grad_buffer import _ParamAndGradBucketGroup
            self.origin_funcs.append(_ParamAndGradBucketGroup.start_grad_sync)
            self.bucket_class = _ParamAndGradBucketGroup
            _ParamAndGradBucketGroup.start_grad_sync = patch_sync(_ParamAndGradBucketGroup.start_grad_sync)
            monitor.enable_megatron = True
            logger.info("megatron version is > core_r0.8.0 <= core_r0.9.0")
        except ImportError:
            monitor.enable_megatron = False | monitor.enable_megatron

    def restore_grad_sync(self, monitor):
        if not monitor.enable_megatron:
            return

        self.bucket_class.start_grad_sync = self.origin_funcs[0]

    def _get_single_state(self, torch_opt):
        state = {}
        if hasattr(torch_opt, 'param_to_cpu_states_map'):
            state = torch_opt.param_to_cpu_states_map
        elif hasattr(torch_opt, 'state'):
            state = torch_opt.state
        elif hasattr(torch_opt, 'optimizer') and hasattr(torch_opt.optimizer, 'state'):
            state = torch_opt.optimizer.state
        self.state.update(state)


class MegatronMixPrecisionOptimizerMon(OptimizerMon):
    """
    混合精度优化器监测类。在混合精度训练中监测和管理优化器。
    混合精度训练通过适当降低某些计算的精度来加速训练过程并减少内存消耗。
    """

    def map_fp16_to_fp32_param(self, torch_opt):
        for fp16_group, fp32_group in zip(torch_opt.float16_groups, torch_opt.fp32_from_float16_groups):
            for fp16_param, fp32_param in zip(fp16_group, fp32_group):
                self.fp16_to_fp32_param[fp16_param] = fp32_param


class MegatronDistributedOptimizerMon(OptimizerMon):
    def map_fp16_to_fp32_param(self, torch_opt):
        if not (hasattr(torch_opt, "model_float16_groups") and
                hasattr(torch_opt, "shard_fp32_from_float16_groups")):
            raise Exception(
                "megatron distributed optimizer should have model_float16_groups and shard_fp32_from_float16_groups, "
                "if not, please check megatron-lm version")
        for fp16_group, shard_fp32_group in zip(torch_opt.model_float16_groups,
                                                torch_opt.shard_fp32_from_float16_groups):
            for fp16_param, shard_fp32_param in zip(fp16_group, shard_fp32_group):
                self.fp16_to_fp32_param[fp16_param] = shard_fp32_param


class MegatronChainedDistributedOptimizerMon(MegatronDistributedOptimizerMon):
    def map_fp16_to_fp32_param(self, torch_opt):
        for opt in torch_opt.chained_optimizers:
            super().map_fp16_to_fp32_param(opt)


class MegatronChainedMixPrecisionOptimizerMon(MegatronMixPrecisionOptimizerMon):
    def map_fp16_to_fp32_param(self, torch_opt):
        for opt in torch_opt.chained_optimizers:
            super().map_fp16_to_fp32_param(opt)


class DeepSpeedZeroOptimizerMon(OptimizerMon):
    """
    Base monitor class for DeepSpeed ZeRO optimizer.
    ZeRO stage 0 no partition
    ZeRO stage 1 partitions optimizer states across data parallel processes.
    ZeRO stage 2 additionally partitions gradients.
    ZeRO stage 3 additionally partitions parameters.

    This class provides monitoring capabilities for ZeRO optimizers by:
    - Handling gradient collection for different ZeRO stages
    - Managing optimizer state access for monitoring
    """

    def __init__(self, torch_opt):
        super().__init__(torch_opt)
        self.stage = ''
        self.bit16_groups = []
        self.fp32_flat_groups = []
        self.param2group = ()
        self.param2index = []
        self.group_offset = {}

    @abstractmethod
    def get_grad_for_param(self, lp_param, group_idx, param_id):
        raise NotImplementedError

    def param_not_in_partition(self, lp_param, group_idx):
        param_slice_mapping = self.torch_opt.state_dict()['param_slice_mappings'][group_idx]
        hp_address = param_slice_mapping.get(self.torch_opt.param_names.get(lp_param))
        return hp_address is None

    def get_position(self, lp_param, group_idx):
        param_slice_mapping = self.torch_opt.state_dict()['param_slice_mappings'][group_idx]
        hp_address = param_slice_mapping.get(self.torch_opt.param_names.get(lp_param))
        return hp_address.start, hp_address.numel

    def get_group_index(self):
        param2group = {}
        for group_idx, bit16_group in enumerate(self.bit16_groups):
            for param in bit16_group:
                param2group[param] = group_idx
        return param2group

    def get_param_index(self, lp_param, group_idx):
        if not self.param2index:
            for group in self.bit16_groups:
                param2index = {}
                for index, param in enumerate(group):
                    param2index[param] = index
                self.param2index.append(param2index)

        return self.param2index[group_idx][lp_param]

    def narrow_from_flatten(self, param, flatten_state):
        if flatten_state is None:
            return flatten_state
        group_idx = self.param2group[param]
        if self.param_not_in_partition(param, group_idx):
            return None
        start, numel = self.get_position(param, group_idx)
        return flatten_state.narrow(0, start, numel)

    def map_fp16_to_fp32_param(self, torch_opt):
        for group_idx, group in enumerate(self.bit16_groups):
            for param in group:
                self.fp16_to_fp32_param[param] = self.fp32_flat_groups[group_idx]

    def fetch_grad(self, monitor, params2name):
        grad_dict = {}
        for lp_param, name in params2name.items():
            group_idx = self.param2group[lp_param]
            param_id = self.get_param_index(lp_param, group_idx)
            if self.param_not_in_partition(lp_param, group_idx):
                continue
            if self.stage == '1or2':
                param_id = param_id - self.group_offset[group_idx] - 1
            grad = self.get_grad_for_param(lp_param, group_idx, param_id)
            tag = monitor.name2tag.get(name, {}).get(MonitorConst.POST_GRAD)
            monitor.register_param_call_id("hook_optimizer", tag)
            grad_dict[tag] = grad

        return grad_dict

    def patch_grad_sync(self, monitor):
        pass

    def restore_grad_sync(self, monitor):
        pass


class DeepSpeedZeroOptimizerStage0Mon(DeepSpeedZeroOptimizerMon):
    def __init__(self, torch_opt):
        super().__init__(torch_opt)
        self.stage = '0'
        self.bit16_groups = torch_opt.bf16_groups
        self.fp32_flat_groups = torch_opt.fp32_groups_flat_partition
        self.param2group = self.get_group_index()

    def get_grad_for_param(self, lp_param, group_idx, param_id):
        return self.torch_opt.fp32_groups_gradient_dict[group_idx][param_id]


class DeepSpeedZeroOptimizerStage1or2Mon(DeepSpeedZeroOptimizerMon):
    def __init__(self, torch_opt):
        super().__init__(torch_opt)
        self.stage = '1or2'
        self.bit16_groups = torch_opt.bit16_groups
        self.fp32_flat_groups = torch_opt.single_partition_of_fp32_groups
        self.param2group = self.get_group_index()
        self.group_offset = {}
        self.get_group_offset()

    def get_grad_for_param(self, lp_param, group_idx, param_id):
        if getattr(self.torch_opt, "cpu_offload", False):
            grads = self.torch_opt.single_partition_of_fp32_groups[group_idx].grad
            start, numel = self.get_position(lp_param, group_idx)
            grad = grads.narrow(0, start, numel)
        else:
            grad = self.torch_opt.averaged_gradients[group_idx][param_id]
        return grad

    def get_group_offset(self):
        for group_idx, group in enumerate(self.bit16_groups):
            self.group_offset[group_idx] = -1
            for lp_param in group:
                if self.param_not_in_partition(lp_param, group_idx):
                    self.group_offset[group_idx] = self.get_param_index(lp_param, group_idx)
                else:
                    break

    def patch_grad_sync(self, monitor):
        def patch_sync(reduce_func):
            def wrapper(zero_optimizer, *args, **kwargs):
                grad_dict = {}
                for i, param, _ in zero_optimizer.params_in_ipg_bucket:
                    if isinstance(param, int):  # for ds >= 0.17.0
                        param = zero_optimizer.bit16_groups[i][param]
                    name = monitor.param2name[param]
                    tag = monitor.name2tag.get(name, {}).get(MonitorConst.PRE_GRAD)
                    grad_dict[tag] = zero_optimizer.get_gradient_for_reduction(param)
                    monitor.register_param_call_id("sync_grad_func", tag)
                get_metrics(monitor.ops, grad_dict, monitor.eps, monitor.grad_context.pre)
                out = reduce_func(zero_optimizer, *args, **kwargs)
                return out

            return wrapper

        try:
            from deepspeed.runtime.zero.stage_1_and_2 import DeepSpeedZeroOptimizer
            self.origin_funcs = [
                DeepSpeedZeroOptimizer.average_tensor,
                DeepSpeedZeroOptimizer.buffered_reduce_fallback
            ]
            DeepSpeedZeroOptimizer.average_tensor = patch_sync(DeepSpeedZeroOptimizer.average_tensor)
            DeepSpeedZeroOptimizer.buffered_reduce_fallback = \
                patch_sync(DeepSpeedZeroOptimizer.buffered_reduce_fallback)
            monitor.enable_deepspeed = True
            logger.info('deepspeed enabled')
        except Exception as e:
            monitor.enable_deepspeed = False | monitor.enable_deepspeed
            logger.warning('Seems using deepspeed zero 1 or 2. But patch average tensor failed')

    def restore_grad_sync(self, monitor):
        if not monitor.enable_deepspeed:
            return

        from deepspeed.runtime.zero.stage_1_and_2 import DeepSpeedZeroOptimizer
        DeepSpeedZeroOptimizer.average_tensor = self.origin_funcs[0]
        DeepSpeedZeroOptimizer.buffered_reduce_fallback = self.origin_funcs[1]


class DeepSpeedZeroOptimizerStage3Mon(DeepSpeedZeroOptimizerMon):
    def __init__(self, torch_opt):
        super().__init__(torch_opt)
        self.stage = '3'
        self.bit16_groups = torch_opt.fp16_groups
        self.fp32_flat_groups = torch_opt.fp32_partitioned_groups_flat
        self.param2group = self.get_group_index()

    def param_not_in_partition(self, lp_param, group_idx):
        """Each param partioned across all zero ranks"""
        return False

    def get_position(self, lp_param, group_idx):
        param_id = self.torch_opt.get_param_id(lp_param)
        return self.torch_opt.grad_position[param_id][1:]

    def get_grad_for_param(self, lp_param, group_idx, param_id):
        return self.torch_opt.averaged_gradients[group_idx][param_id]


class OptimizerMonFactory:
    _optimizer_mon_map = {
        "FP32Optimizer": OptimizerMon,
        "Float16OptimizerWithFloat16Params": MegatronMixPrecisionOptimizerMon,
        "DistributedOptimizer": MegatronDistributedOptimizerMon,
        "SwapDistributedOptimizer": MegatronDistributedOptimizerMon,
        "ChainedDistributedOptimizer": MegatronChainedDistributedOptimizerMon,
        "ChainedSwapDistributedOptimizer": MegatronChainedDistributedOptimizerMon,
        "ChainedFloat16OptimizerWithFloat16Params": MegatronChainedMixPrecisionOptimizerMon,
        "BF16_Optimizer": DeepSpeedZeroOptimizerStage0Mon,
        "DeepSpeedZeroOptimizer": DeepSpeedZeroOptimizerStage1or2Mon,
        "DeepSpeedZeroOptimizer_Stage3": DeepSpeedZeroOptimizerStage3Mon,
        "Adam": OptimizerMon
    }

    @staticmethod
    def create_optimizer_mon(optimizer):
        # auto replace opt_ty
        optimizer_class = optimizer.__class__.__name__
        if optimizer_class == "ChainedOptimizer":
            optimizer_class = "Chained" + optimizer.chained_optimizers[0].__class__.__name__
        logger.info(f'The optimizer type is {optimizer_class}')

        optimizer_mon_class = OptimizerMonFactory._optimizer_mon_map.get(optimizer_class, OptimizerMon)
        return optimizer_mon_class(optimizer)
